import numpy as np
from psyneulink.core.scheduling.time import TimeScale
from psyneulink.core.globals.utilities import NodeRole
from psyneulink.core.components.functions.transferfunctions import Linear, Logistic, ReLU
from psyneulink.core.globals.context import Context, ContextFlags, handle_external_context
from psyneulink.core import llvm as pnlvm
from psyneulink.library.compositions.compiledoptimizer import AdamOptimizer,SGDOptimizer
from psyneulink.library.compositions.compiledloss import MSELoss
import ctypes
from collections import deque

debug_env = pnlvm.debug_env

try:
    import torch
    from torch import nn
    torch_available = True
except ImportError:
    torch_available = False


__all__ = ['PytorchModelCreator']
# Class that is called to create pytorch representations of autodiff compositions based on their processing graphs.
# Called to do so when the composition is run for the first time.

# Note on notation: the "nodes" that are constantly referred to are vertices of the composition's processing
# graph. For general compositions, the component a node represents can be a mechanism or a nested composition,
# but for autodiff compositions, nodes always represent mechanisms. "Nodes" can be thought of as
# (but are not literally) mechanisms.


class PytorchModelCreator(torch.nn.Module):

    # sets up parameters of model & the information required for forward computation
    def __init__(self, processing_graph, param_init_from_pnl, execution_sets, device, context=None, composition=None):

        if not torch_available:
            raise Exception('Pytorch python module (torch) is not installed. Please install it with '
                            '`pip install torch` or `pip3 install torch`')

        super(PytorchModelCreator, self).__init__()

        self.execution_sets = execution_sets  # saved for use in the forward method
        # dict mapping PNL nodes to their forward computation information
        self.component_to_forward_info = {}
        # dict mapping PNL projections to Pytorch weights
        self.projections_to_pytorch_weights = {}
        # dict mapping PNL mechanisms to Pytorch biases
        self.mechanisms_to_pytorch_biases = {}
        # list that Pytorch optimizers will use to keep track of parameters
        self.params = nn.ParameterList()
        self.device = device
        self.__bin_exec_func = None
        self._forward_llvm_func = None
        self._cached_param_list = None
        self._cached_tupleized_param_list = None

        self._composition = composition

        for i, current_exec_set in enumerate(self.execution_sets):
            # SKG: We have to add a counter to map components to an internal int id (for bin execute purposes, since there is no concept of a 'dict' in llvm)
            id_map_ct = 0
            for component in current_exec_set:
                id_map_ct += 1

                value = None  # the node's (its mechanism's) value
                biases = None  # the node's bias parameters
                function = self.function_creator(
                    component, context)  # the node's function
                afferents = {}  # dict for keeping track of afferent nodes and their connecting weights
                if param_init_from_pnl:
                    if component.parameters.value._get(context) is None:
                        value = torch.tensor(component.parameters.value.get(None)[0], device=self.device)
                    else:
                        value = torch.tensor(component.parameters.value._get(context)[0], device=self.device)
                else:
                    input_length = len(
                        component.input_states[0].parameters.value.get(None))
                    value = torch.zeros(
                        input_length, device=self.device).double()

                # if `node` is not an origin node (origin nodes don't have biases or afferent connections)
                if i != 0:
                    # if not copying parameters from psyneulink, set up pytorch biases for node
                    if not param_init_from_pnl:
                        input_length = len(
                            component.input_states[0].parameters.value.get(None))
                        biases = nn.Parameter(torch.zeros(
                            input_length, device=self.device).double())
                        self.params.append(biases)
                        self.mechanisms_to_pytorch_biases[component] = biases
                    # iterate over incoming projections and set up pytorch weights for them
                    for mapping_proj in component.path_afferents:

                        # get projection, sender node--pdb for projection
                        input_component = mapping_proj.sender.owner
                        input_node = processing_graph.comp_to_vertex[input_component]

                        # CW 12/3/18: Check this logic later
                        proj_matrix = mapping_proj.parameters.matrix._get(
                            context)
                        if proj_matrix is None:
                            proj_matrix = mapping_proj.parameters.matrix.get(
                                None)
                        # set up pytorch weights that correspond to projection. If copying params from psyneulink,
                        # copy weight values from projection. Otherwise, use random values.
                        if param_init_from_pnl:
                            weights = nn.Parameter(torch.tensor(
                                proj_matrix.copy(), device=self.device).double())
                        else:
                            weights = nn.Parameter(torch.rand(
                                np.shape(proj_matrix), device=self.device).double())
                        afferents[input_node] = weights
                        self.params.append(weights)
                        self.projections_to_pytorch_weights[mapping_proj] = weights
                node_forward_info = [value, biases, function, afferents, component]

                self.component_to_forward_info[component] = node_forward_info

        # CW 12/3/18: this copies by reference so in theory it only needs to be called during init
        # but we call copy_weights_to_psyneulink after every run in order to make Autodiff less stateful
        self.copy_weights_to_psyneulink(context)


    # gets the index of 'afferent_node' in the forward info weights list
    def _get_afferent_node_index(self,node,afferent_node):
        forward_info_weights = self.component_to_forward_info[node][3]
        for (idx,vertex) in enumerate(forward_info_weights):
            if vertex.component == afferent_node:
                return idx

    # returns a list of all efferent nodes and weights stored in component_to_forward_info
    def _get_afferent_nodes(self,node):
        forward_info_weights = self.component_to_forward_info[node][3]
        return [(vertex.component,weights) for (vertex,weights) in forward_info_weights.items()]

    # defines input type
    def _get_input_struct_type(self, ctx):  # Test case: {[1 x [2 x double]]}
        input_ty = [None]*len(self.execution_sets[0])
        for component in self.execution_sets[0]:
            component_id = self._composition._get_node_index(component)
            input_ty[component_id] = ctx.convert_python_struct_to_llvm_ir(
                component.defaults.variable[0])
        struct_ty = pnlvm.ir.types.LiteralStructType(input_ty)
        return struct_ty

    def _get_data_struct_type(self, ctx):
        # Ensures that data struct is the same as the autodiffcomp
        return self._composition._get_data_struct_type(ctx)

    # gets the params (i.e. weights and biases) structure for llvm
    # param struct shape:
    #   For weights:    [component_id][0][efferent_id_map_id]
    #   For bias:       [component_id][1]
    # The efferent id map is needed for weights because we must start from index 0, so each node keeps track of its efferents by iterating them
    def _get_param_struct_type(self, ctx):

        param_list = [None]*len(self._composition.nodes)
        for (node, forward_info) in self.component_to_forward_info.items():

            node_idx = self._composition._get_node_index(node)

            # 1) setup weights
            afferents = forward_info[3]
            weight_array = [None]*len(afferents)
            for (afferent_vertex, weights) in afferents.items():
                afferent_node = afferent_vertex.component
                afferent_index = self._get_afferent_node_index(node,afferent_node)
                if "no_ref_pass" not in debug_env:
                    afferent_weight = pnlvm.ir.types.IntType(64)
                else:
                    afferent_weight = ctx.convert_python_struct_to_llvm_ir(
                        weights.detach().numpy())
                weight_array[afferent_index] = afferent_weight
            node_params = [pnlvm.ir.types.LiteralStructType(weight_array)]

            # 2) setup bias
            bias = forward_info[1]
            if bias is not None:
                if "no_ref_pass" not in debug_env:
                    # FIXME: This should use a pointer type
                    node_params.append(pnlvm.ir.types.IntType(64))
                else:
                    node_params.append(ctx.convert_python_struct_to_llvm_ir(
                        bias.detach().numpy()))

            param_list[node_idx] = pnlvm.ir.types.LiteralStructType(node_params)
        return pnlvm.ir.types.LiteralStructType(param_list)

    def _get_param_initializer(self):
        if self._cached_param_list is None:
            param_list = [None]*len(self._composition.nodes)
            for (node, forward_info) in self.component_to_forward_info.items():

                node_idx = self._composition._get_node_index(node)

                # 1) initialize weights
                afferents = forward_info[3]
                weight_array = [None]*len(afferents)
                for (afferent_vertex, weights) in afferents.items():
                    afferent_node = afferent_vertex.component
                    afferent_index = self._get_afferent_node_index(node,afferent_node)
                    if "no_ref_pass" not in debug_env: # this gets the actual memory address of the weights - is static (according to https://github.com/numpy/numpy/issues/13906)
                        afferent_weight = weights.detach().numpy().ctypes.data
                    else:
                        afferent_weight = weights.detach().numpy()
                    weight_array[afferent_index] = afferent_weight
                node_params = [weight_array]

                # 2) initialize bias
                bias = forward_info[1]
                if bias is not None:
                    if "no_ref_pass" not in debug_env:
                        node_params += [bias.detach().numpy().ctypes.data]
                    else:
                        node_params.append(bias.detach().numpy())

                param_list[node_idx] = node_params
            if "no_ref_pass" not in debug_env:
                self._cached_param_list = pnlvm.execution._tupleize(param_list)
            else:
                self._cached_param_list = param_list
        if "no_ref_pass" not in debug_env:
            return self._cached_param_list
        return pnlvm.execution._tupleize(self._cached_param_list)

    def _get_state_struct_type(self, ctx):
        return self._composition._get_state_struct_type(ctx)

    # generates llvm function for self.forward
    def _gen_llvm_function(self, extra_args=[], name=None):
        llvm_func = None
        with pnlvm.LLVMBuilderContext.get_global() as ctx:
            args = [ctx.get_state_struct_type(self).as_pointer(),
                    ctx.get_param_struct_type(self).as_pointer(),
                    ctx.get_input_struct_type(self).as_pointer(),
                    ctx.get_data_struct_type(self).as_pointer()
                    ]
            builder = ctx.create_llvm_function(args+extra_args, self, name)
            llvm_func = builder.function

            context, params, arg_in, arg_out = llvm_func.args[:len(args)]
            self._gen_llvm_forward_function_body(
                ctx, builder, context, params, arg_in, arg_out)
            builder.ret_void()
            llvm_func = builder.function
        self._forward_llvm_func = llvm_func
        return llvm_func

    #FIXME: Move _gen functions to helper or change builtins to directly accept aggregate types
    def _gen_inject_bin_function_call(self, ctx, builder, bin_func, vector, output_vec=None):
        dim = len(vector.type.pointee)
        if output_vec is None:
            output_vec = builder.alloca(pnlvm.ir.types.ArrayType(ctx.float_ty, dim))
        assert len(output_vec.type.pointee) == dim

        # Get the pointer to the first element of the array to convert from [? x double]* -> double*
        vec_in = builder.gep(vector, [ctx.int32_ty(0), ctx.int32_ty(0)])
        vec_out = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

        builder.call(bin_func, [vec_in, ctx.int32_ty(dim), vec_out])
        return output_vec

    def _gen_inject_vec_copy(self, ctx, builder, vector, output_vec=None):
        dim = len(vector.type.pointee)
        if output_vec is None:
            output_vec = builder.alloca(pnlvm.ir.types.ArrayType(ctx.float_ty, dim))
        assert len(output_vec.type.pointee) == dim

        # Get the pointer to the first element of the array to convert from [? x double]* -> double*
        vec_in = builder.gep(vector, [ctx.int32_ty(0), ctx.int32_ty(0)])
        vec_out = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

        builtin = ctx.get_llvm_function("__pnl_builtin_vec_copy")
        builder.call(builtin, [vec_in, ctx.int32_ty(dim), vec_out])
        return output_vec

    def _gen_inject_vec_binop(self, ctx, builder, op, u, v, output_vec=None):
        dim = len(u.type.pointee)
        assert len(v.type.pointee) == dim
        if output_vec is None:
            output_vec = builder.alloca(pnlvm.ir.types.ArrayType(ctx.float_ty, dim))
        assert len(output_vec.type.pointee) == dim

        # Get the pointer to the first element of the array to convert from [? x double]* -> double*
        vec_u = builder.gep(u, [ctx.int32_ty(0), ctx.int32_ty(0)])
        vec_v = builder.gep(v, [ctx.int32_ty(0), ctx.int32_ty(0)])
        vec_out = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

        builder.call(ctx.get_llvm_function(op), [vec_u, vec_v, ctx.int32_ty(dim), vec_out])
        return output_vec

    def _gen_inject_vec_add(self, ctx, builder, u, v, output_vec=None):
        return self._gen_inject_vec_binop(ctx, builder, "__pnl_builtin_vec_add", u, v, output_vec)

    def _gen_inject_vec_sub(self, ctx, builder, u, v, output_vec=None):
        return self._gen_inject_vec_binop(ctx, builder, "__pnl_builtin_vec_sub", u, v, output_vec)

    def _gen_inject_vec_hadamard(self, ctx, builder, u ,v, output_vec=None):
        return self._gen_inject_vec_binop(ctx, builder, "__pnl_builtin_vec_hadamard", u, v, output_vec)

    def _gen_inject_mat_binop(self, ctx, builder, op, m1, m2, output_mat=None):
        x = len(m1.type.pointee)
        y = len(m1.type.pointee.element)
        assert len(m2.type.pointee) == x and len(m2.type.pointee.element) == y

        if output_mat is None:
            output_mat = builder.alloca(
                pnlvm.ir.types.ArrayType(
                    pnlvm.ir.types.ArrayType(ctx.float_ty, y), x))
        assert len(output_mat.type.pointee) == x
        assert len(output_mat.type.pointee.element) == y

        builtin = ctx.get_llvm_function(op)
        builder.call(builtin, [builder.bitcast(m1, ctx.float_ty.as_pointer()),
                               builder.bitcast(m2, ctx.float_ty.as_pointer()),
                               ctx.int32_ty(x), ctx.int32_ty(y),
                               builder.bitcast(output_mat, ctx.float_ty.as_pointer())])
        return output_mat

    def _gen_inject_mat_add(self, ctx, builder, m1, m2, output_mat=None):
        return self._gen_inject_mat_binop(ctx, builder, "__pnl_builtin_mat_add", m1, m2, output_mat)

    def _gen_inject_mat_sub(self, ctx, builder, m1, m2, output_mat=None):
        return self._gen_inject_mat_binop(ctx, builder, "__pnl_builtin_mat_sub", m1, m2, output_mat)

    def _gen_inject_mat_hadamard(self, ctx, builder, m1, m2, output_mat=None):
        return self._gen_inject_mat_binop(ctx, builder, "__pnl_builtin_mat_hadamard", m1, m2, output_mat)

    def _gen_inject_mat_scalar_mult(self, ctx, builder, m1, s, output_mat=None):
        x = len(m1.type.pointee)
        y = len(m1.type.pointee.element)
        if output_mat is None:
            output_mat = builder.alloca(
                pnlvm.ir.types.ArrayType(
                    pnlvm.ir.types.ArrayType(ctx.float_ty, y), x))
        assert len(output_mat.type.pointee) == x
        assert len(output_mat.type.pointee.element) == y

        builtin = ctx.get_llvm_function("__pnl_builtin_mat_scalar_mult")
        builder.call(builtin, [builder.bitcast(m1, ctx.float_ty.as_pointer()),
                               s, ctx.int32_ty(x), ctx.int32_ty(y),
                               builder.bitcast(output_mat, ctx.float_ty.as_pointer())])
        return output_mat

    def _gen_inject_vxm(self, ctx, builder, m1, m2, output_vec=None):
        y = len(m2.type.pointee)
        z = len(m2.type.pointee.element)
        assert len(m1.type.pointee) == y
        # create output vec
        if output_vec is None:
            output_vec = builder.alloca(pnlvm.ir.types.ArrayType(ctx.float_ty, z))
        assert len(output_vec.type.pointee) == z

        # Get the pointer to the first element of the array to convert from [? x double]* -> double*
        v = builder.gep(m1, [ctx.int32_ty(0), ctx.int32_ty(0)])
        out = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

        builtin = ctx.get_llvm_function("__pnl_builtin_vxm")
        builder.call(builtin, [v, builder.bitcast(m2, ctx.float_ty.as_pointer()),
                               ctx.int32_ty(y), ctx.int32_ty(z), out])
        return output_vec

    def _gen_inject_vxm_transposed(self, ctx, builder, m1, m2, output_vec=None):
        y = len(m2.type.pointee)
        z = len(m2.type.pointee.element)
        assert len(m1.type.pointee) == z
        # create output vec
        if output_vec is None:
            output_vec = builder.alloca(pnlvm.ir.types.ArrayType(ctx.float_ty, y))
        assert len(output_vec.type.pointee) == y

        # Get the pointer to the first element of the array to convert from [? x double]* -> double*
        v = builder.gep(m1, [ctx.int32_ty(0), ctx.int32_ty(0)])
        out = builder.gep(output_vec, [ctx.int32_ty(0), ctx.int32_ty(0)])

        builtin = ctx.get_llvm_function("__pnl_builtin_vxm_transposed")
        builder.call(builtin, [v, builder.bitcast(m2, ctx.float_ty.as_pointer()),
                               ctx.int32_ty(y), ctx.int32_ty(z), out])
        return output_vec

    # gets a pointer for the weights matrix between node and afferent_node
    def _gen_get_node_weight_ptr(self, ctx, builder,model_params,node,afferent_node):
        node_idx = self._composition._get_node_index(node)
        forward_info_weights = self.component_to_forward_info[node][3]
        afferent_node_index = self._get_afferent_node_index(node,afferent_node)
        for (vertex,matrix) in forward_info_weights.items():
            if vertex.component == afferent_node:
                weight_matrix = matrix
                break
        dim_x,dim_y = weight_matrix.detach().numpy().shape
        node_weights = builder.gep(model_params,[ctx.int32_ty(0),
                                                ctx.int32_ty(node_idx),
                                                ctx.int32_ty(0),
                                                ctx.int32_ty(afferent_node_index)])
        if "no_ref_pass" not in debug_env:
            mem_addr = builder.load(node_weights)
            node_weights = builder.inttoptr(mem_addr, pnlvm.ir.types.ArrayType(
                pnlvm.ir.types.ArrayType(ctx.float_ty, dim_y), dim_x).as_pointer())

        return node_weights,dim_x,dim_y

    def _gen_llvm_forward_function_body(self, ctx, builder, _, params, arg_in, arg_out, store_z_values=False):
        out_t = arg_out.type.pointee
        if isinstance(out_t, pnlvm.ir.ArrayType) and isinstance(out_t.element, pnlvm.ir.ArrayType):
            assert len(out_t) == 1

        if store_z_values is True:
            z_values = {}
        for i, current_exec_set in enumerate(self.execution_sets):

            for component in current_exec_set:
                component_id = self._composition._get_node_index(component)
                biases = self.component_to_forward_info[component][1]
                value = self._get_output_value_ptr(
                    ctx, builder, arg_out, component_id)
                afferents = self.component_to_forward_info[component][3]

                if i == 0:
                    cmp_arg = builder.gep(arg_in, [ctx.int32_ty(0), ctx.int32_ty(component_id)])
                else:
                    # is_set keeps track of if we already have valid (i.e. non-garbage) values inside the alloc'd value
                    is_set = False
                    for input_vertex, weights in afferents.items():
                        input_node = input_vertex.component
                        ctx.inject_printf(builder,f"COMPILED FORWARD {input_node} -> {component}\n")
                        input_node_idx = self._composition._get_node_index(
                            input_node)
                        # frozen_values[input_node.component]
                        input_value = self._get_output_value_ptr(
                            ctx, builder, arg_out, input_node_idx)

                        # We cast the ctype weights array to llvmlite pointer
                        weights_llvmlite, _, _ = self._gen_get_node_weight_ptr(ctx, builder, params, component, input_node)
                        weighted_inp = self._gen_inject_vxm(ctx, builder, input_value, weights_llvmlite)
                        if is_set == False:
                            # copy weighted_inp to value
                            self._gen_inject_vec_copy(ctx, builder, weighted_inp, value)
                            is_set = True
                        else:
                            # add to value
                            self._gen_inject_vec_add(ctx, builder, weighted_inp, value, value)

                    cmp_arg = value

                # Apply Activation Func to values
                if store_z_values is True:
                    z_values[component] = self._gen_inject_vec_copy(ctx, builder, cmp_arg)
                bin_func = ctx.get_llvm_function(self.bin_function_creator(ctx,component).name)
                self._gen_inject_bin_function_call(ctx, builder, bin_func, cmp_arg, value)

                # TODO: Add bias to value
                # if biases is not None:
                #   value = value + biases
                if store_z_values is True:
                    ctx.inject_printf_float_array(builder, z_values[component], prefix=f"Z VALUE FOR {component} :\t")
                ctx.inject_printf_float_array(builder, value, prefix=f"FORWARD VALUE FOR {component} :\t")
        if store_z_values is True:
            return z_values

    # generates a function responsible for a single epoch of the training
    def _gen_llvm_training_backprop(self, ctx, optimizer, loss, extra_args=[]):
        composition = self._composition
        learning_targets = pnlvm.ir.LiteralStructType([
            ctx.int32_ty,  # dimensionality
            pnlvm.ir.IntType(64)
        ])
        args = [ctx.get_state_struct_type(self).as_pointer(),
                ctx.get_param_struct_type(self).as_pointer(),
                ctx.get_input_struct_type(self).as_pointer(),
                ctx.get_data_struct_type(self).as_pointer(),
                optimizer._get_optimizer_struct_type(ctx).as_pointer(),
                learning_targets.as_pointer(),  # inputs
                learning_targets.as_pointer(),  # targets
                ctx.int32_ty  # input idx
                ]
        name = self._composition.name+"_training_backprop"
        builder = ctx.create_llvm_function(args+extra_args, self, name)
        llvm_func = builder.function
        for a in llvm_func.args:
            if isinstance(a.type, pnlvm.ir.PointerType):
                a.attributes.add('noalias')

        model_context, model_params, model_input, model_output, optim_struct, input_struct_ptr, target_struct_ptr, input_idx = llvm_func.args[:len(args)]
        # setup useful mappings
        input_nodes = composition.get_nodes_by_role(NodeRole.INPUT)
        output_nodes = composition.get_nodes_by_role(NodeRole.OUTPUT)

        def _get_node_array_ptr(node, node_idx, struct_ptr):
            array_ptr = builder.gep(struct_ptr, [ctx.int32_ty(node_idx), ctx.int32_ty(1)])
            array_ptr = builder.load(array_ptr)
            array_ty = pnlvm.ir.ArrayType(pnlvm.ir.ArrayType(ctx.float_ty, len(node.defaults.value[0])), 1)
            return builder.inttoptr(array_ptr, array_ty.as_pointer())

        node_input_arrays = {node: _get_node_array_ptr(node, i, input_struct_ptr) for i, node in enumerate(input_nodes)}

        node_target_arrays = {node: _get_node_array_ptr(node, i, target_struct_ptr) for i, node in enumerate(output_nodes)}


        # initialize optimizer params:
        delta_w = builder.gep(optim_struct,[ctx.int32_ty(0),ctx.int32_ty(optimizer._DELTA_W_NUM)])

        # first we copy input values to data struct of input_CIM
        for node in input_nodes:
            node_idx = composition._get_node_index(node)
            node_input_array_ptr = builder.gep(node_input_arrays[node],
                                               [input_idx, ctx.int32_ty(0)])
            node_model_input = builder.gep(model_input,[ctx.int32_ty(0), ctx.int32_ty(node_idx)])
            self._gen_inject_vec_copy(ctx, builder, node_input_array_ptr, node_model_input)

            ctx.inject_printf_float_array(builder, node_input_array_ptr, prefix=f"\tNODE {node_idx} INPUT: ")

        # 2) call forward computation
        z_values = self._gen_llvm_forward_function_body(
            ctx, builder, model_context, model_params, model_input, model_output, store_z_values=True)
        # 3) compute errors

        ctx.inject_printf(builder, "\tCOMPUTE ERR FOR INPUT %d\n", input_idx)

        error_dict = {}
        backprop_queue = deque()
        for node in output_nodes:
            backprop_queue.append(node)

        loss_fn = ctx.get_llvm_function(loss._gen_call_function(ctx).name)
        total_loss = builder.alloca(ctx.float_ty)
        builder.store(ctx.float_ty(0),total_loss)

        while(len(backprop_queue) > 0):
            node = backprop_queue.popleft()
            if node in error_dict or not hasattr(node, "afferents") or node == composition.input_CIM or node in input_nodes:
                continue

            for (afferent_node,weights) in self._get_afferent_nodes(node):
                backprop_queue.append(afferent_node)

            node_idx = composition._get_node_index(node)


            activation_func_derivative_bin_func = ctx.get_llvm_function(self.bin_function_derivative_creator(ctx,node).name)
            activation_func_derivative = self._gen_inject_bin_function_call(ctx, builder, activation_func_derivative_bin_func, z_values[node])

            error_val = builder.alloca(z_values[node].type.pointee)

            error_dict[node] = error_val

            if node in output_nodes:
                # We handle output layer here
                # compute  dC/da = a_l - y(x) (TODO: Allow other cost functions! This only applies to MSE)
                node_target = builder.gep(node_target_arrays[node], [
                                            input_idx, ctx.int32_ty(0)])
                node_output = self._get_output_value_ptr(ctx,builder,model_output,node_idx)

                tmp_loss = loss._gen_inject_lossfunc_call(ctx, builder, loss_fn, node_output, node_target)

                ctx.inject_printf_float_array(builder, node_output, override_debug=False)

                ctx.inject_printf(builder,f"tmp loss for {node} :%f\n",tmp_loss,override_debug=False)
                builder.store(builder.fadd(builder.load(total_loss),tmp_loss),total_loss)
                loss_derivative = loss._gen_inject_loss_differential(ctx, builder, node_output, node_target)
                # compute δ_l = dσ/da ⊙ σ'(z)

                self._gen_inject_vec_hadamard(ctx, builder, activation_func_derivative, loss_derivative, error_val)

            else:
                # We propagate error backwards from next layer

                is_set = False

                # We calculate δ_(l-1) = sum (a_(l-1) W^T) ⊙ δ_l, where (l-1) is the current layer, l is layer of efferents, summed over all efferents
                efferents = [
                    proj.receiver._owner for proj in node.efferents]
                for efferent_node in efferents:
                    efferent_node_error = error_dict[efferent_node]

                    weights_llvmlite, _, _ = self._gen_get_node_weight_ptr(ctx, builder, model_params, efferent_node, node)

                    if is_set is False:
                        self._gen_inject_vxm_transposed(ctx, builder, efferent_node_error, weights_llvmlite, error_val)
                        is_set = True
                    else:
                        new_val = self._gen_inject_vxm_transposed(ctx, builder, efferent_node_error, weights_llvmlite)

                        self._gen_inject_vec_add(ctx, builder, new_val, error_val, error_val)

                self._gen_inject_vec_hadamard(ctx, builder, activation_func_derivative, error_val, error_val)

            ctx.inject_printf_float_array(builder, activation_func_derivative, prefix=f"dSIGMA VALUE FOR {node}:\t")
            ctx.inject_printf_float_array(builder, error_val, prefix=f"ERROR VALUE FOR {node}:\t")

        # 4) compute weight gradients
        for (node, err_val) in error_dict.items():
            if node in input_nodes:
                continue
            node_idx = self._composition._get_node_index(node)
            for (afferent_node,weight) in self._get_afferent_nodes(node):
                # get a_(l-1)
                afferent_node_idx = self._get_afferent_node_index(node,afferent_node)

                afferent_node_activation = self._get_output_value_ptr(ctx,builder,model_output,self._composition._get_node_index(afferent_node))

                # get dimensions of weight matrix
                _,weights_dim_x,weights_dim_y = self._gen_get_node_weight_ptr(ctx,builder,model_params,node,afferent_node)
                # update delta_W
                node_delta_w = builder.gep(delta_w,[ctx.int32_ty(0),ctx.int32_ty(node_idx), ctx.int32_ty(afferent_node_idx)])

                with pnlvm.helpers.for_loop_zero_inc(builder, ctx.int32_ty(weights_dim_x), "weight_update_loop_outer") as (b1, weight_row):
                    with pnlvm.helpers.for_loop_zero_inc(b1, ctx.int32_ty(weights_dim_y), "weight_update_loop_inner") as (b2, weight_column):
                        a_val = b2.load(b2.gep(afferent_node_activation,
                                               [ctx.int32_ty(0), weight_row]))
                        d_val = b2.load(b2.gep(err_val,
                                               [ctx.int32_ty(0), weight_column]))
                        old_val = b2.load(b2.gep(node_delta_w,
                                                 [ctx.int32_ty(0), weight_row, weight_column]))
                        new_val = b2.fadd(old_val, b2.fmul(a_val, d_val))
                        b2.store(new_val, b2.gep(node_delta_w,
                                                 [ctx.int32_ty(0), weight_row, weight_column]))

        builder.store(builder.fmul(ctx.float_ty(.5),builder.load(total_loss)),total_loss)
        ctx.inject_printf(builder,"TOTAL LOSS: %f\n",builder.load(total_loss),override_debug=False)
        builder.ret_void()

        return builder.function

    def _gen_llvm_training_function_body(self, ctx, builder, context, params, comp_in, data_arg, cond):
        # 1) Setup autodiff learning stuff
        # if "ref_pass" not in debug_env:
        #    raise Exception("ref_pass must be enabled in debug!")
        # gets a reference to the autodiff_stimuli_struct from params
        composition = self._composition
        autodiff_stimuli_struct = builder.gep(
            params, [ctx.int32_ty(0), ctx.int32_ty(3)])

        epochs = builder.load(builder.gep(autodiff_stimuli_struct, [
                              ctx.int32_ty(0), ctx.int32_ty(0)]))
        num_inputs = builder.load(builder.gep(autodiff_stimuli_struct, [
            ctx.int32_ty(0), ctx.int32_ty(1)]))

        num_target_structs = builder.load(builder.gep(
            autodiff_stimuli_struct, [ctx.int32_ty(0), ctx.int32_ty(2)]))
        # Get pointer to the first element
        target_struct_ptr = builder.gep(
            autodiff_stimuli_struct, [ctx.int32_ty(0), ctx.int32_ty(3), ctx.int32_ty(0)])

        num_input_structs = builder.load(builder.gep(
            autodiff_stimuli_struct, [ctx.int32_ty(0), ctx.int32_ty(4)]))
        # Get pointer to the first element
        input_struct_ptr = builder.gep(
            autodiff_stimuli_struct, [ctx.int32_ty(0), ctx.int32_ty(5), ctx.int32_ty(0)])

        if "const_params" in debug_env:
            const_params = params.type.pointee(
                composition._get_param_initializer(None))
            params = builder.alloca(const_params.type, name="const_params_loc")
            builder.store(const_params, params)

        if "alloca_data" in debug_env:
            data = builder.alloca(data_arg.type.pointee)
            data_vals = builder.load(data_arg)
            builder.store(data_vals, data)
        else:
            data = data_arg

        input_cim_idx = composition._get_node_index(composition.input_CIM)
        model_context = context
        model_params = builder.gep(params, [ctx.int32_ty(0),
                                            ctx.int32_ty(2)])

        # Extract the input that should be inserted into the model
        model_input = builder.gep(data, [ctx.int32_ty(0),
                                         ctx.int32_ty(0),
                                         ctx.int32_ty(input_cim_idx)])
        model_output = builder.gep(data, [ctx.int32_ty(0),
                                          ])

        # setup optimizer
        optimizer_type = self._composition.optimizer_type
        if optimizer_type == 'adam':
            optimizer = AdamOptimizer(self,lr = self._composition.learning_rate)
        elif optimizer_type == 'sgd':
            optimizer = SGDOptimizer(self,lr = self._composition.learning_rate)
        else:
            raise Exception("OPTIMIZER TYPE",optimizer_type,"NOT SUPPORTED")

        # setup loss
        loss_type = self._composition.loss_spec
        if loss_type == 'mse':
            loss = MSELoss(self)
        else:
            raise Exception("LOSS TYPE",loss_type,"NOT SUPPORTED")

        optimizer_struct = builder.alloca(optimizer._get_optimizer_struct_type(ctx))
        optimizer.initialize_optimizer_struct(ctx,builder,optimizer_struct)
        backprop = ctx.get_llvm_function(self._gen_llvm_training_backprop(ctx,optimizer,loss).name)
        optimizer_step = ctx.get_llvm_function(optimizer.step(ctx).name)

        with pnlvm.helpers.for_loop_zero_inc(builder, epochs, "epoch_loop") as (b1, epoch_idx):
            ctx.inject_printf(builder, "\033[0;32mEPOCH %d\033[0m\n", epoch_idx)
            with pnlvm.helpers.for_loop_zero_inc(b1, num_inputs, "input_loop") as (b2, input_idx):
                ctx.inject_printf(b2, "\n\033[0;31mINPUT %d\033[0m\n", input_idx)
                ctx.inject_printf(b2, "OPTIMIZER ZERO GRAD %d\n", input_idx)
                optimizer.zero_grad(ctx,b2,optimizer_struct)
                ctx.inject_printf(b2, "BACKPROP %d\n", input_idx)
                b2.call(backprop,[model_context, model_params, model_input, model_output, optimizer_struct, input_struct_ptr, target_struct_ptr, input_idx])
                ctx.inject_printf(b2, "OPTIMIZER STEP %d\n", input_idx)
                b2.call(optimizer_step,[optimizer_struct,model_params])

    def _get_output_value_ptr(self, ctx, builder, arg_out, index):
        return builder.gep(arg_out, [ctx.int32_ty(0), ctx.int32_ty(0), ctx.int32_ty(index), ctx.int32_ty(0)])

    # performs forward computation for the model
    @handle_external_context()
    def forward(self, inputs, context=None, do_logging=True, scheduler=None):
        outputs = {}  # dict for storing values of terminal (output) nodes

        for i, current_exec_set in enumerate(self.execution_sets):
            frozen_values = {}
            for component in current_exec_set:
                frozen_values[component] = self.component_to_forward_info[component][0]
            for component in current_exec_set:

                # get forward computation info for current component
                biases = self.component_to_forward_info[component][1]
                function = self.component_to_forward_info[component][2]
                afferents = self.component_to_forward_info[component][3]

                # forward computation if we have origin node
                if i == 0:
                    value = function(inputs[component])
                # forward computation if we do not have origin node
                else:
                    value = torch.zeros(
                        len(component.input_states[0].defaults.value), device=self.device).double()
                    for input_node, weights in afferents.items():
                        if input_node.component in current_exec_set:
                            input_value = frozen_values[input_node.component]
                        else:
                            input_value = self.component_to_forward_info[input_node.component][0]
                        value += torch.matmul(input_value, weights)
                    if biases is not None:
                        value = value + biases
                    value = function(value)

                # store the current value of the node
                self.component_to_forward_info[component][0] = value
                if do_logging:
                    old_source = context.source
                    context.source = ContextFlags.COMMAND_LINE
                    detached_value = value.detach().cpu().numpy()
                    component.parameters.value._log_value(detached_value, context)
                    context.source = old_source

                # save value in output list if we're at a node in the last execution set
                if i == len(self.execution_sets) - 1:
                    outputs[component] = value

            if scheduler is not None:
                scheduler.get_clock(context)._increment_time(
                    TimeScale.TIME_STEP)

        # Maybe need to comment this out!
        self.copy_outputs_to_psyneulink(outputs, context)

        if do_logging:
            old_source = context.source
            context.source = ContextFlags.COMMAND_LINE
            self.log_weights(context)
            self.copy_outputs_to_psyneulink(outputs, context)
            context.source = old_source

        return outputs

    def detach_all(self):
        for component, info in self.component_to_forward_info.items():
            info[0].detach_()
            if info[1] is not None:
                info[1].detach_()

    def copy_weights_to_psyneulink(self, context=None):
        for projection, weights in self.projections_to_pytorch_weights.items():
            projection.parameters.matrix._set(
                weights.detach().cpu().numpy(), context)

    def copy_outputs_to_psyneulink(self, outputs, context=None):
        for component, value in outputs.items():
            detached_value = value.detach().cpu().numpy()
            component.parameters.value._set(
                detached_value, context, skip_history=True, skip_log=True)
            component.output_state.parameters.value._set(
                detached_value, context, skip_history=True, skip_log=True)

    @handle_external_context()
    def log_weights(self, context=None):
        for projection, weights in self.projections_to_pytorch_weights.items():
            projection.parameters.matrix._log_value(
                weights.detach().cpu().numpy(), context)

    # Helper method that functions the same as function_creator, but instead injects the computation to the builder
    # FIXME: Change to directly using compiled function methods
    @handle_external_context()
    def bin_function_creator(self, ctx, node, context=None):
        # first try to get cached func
        name = node.name+"_"+node.function.name
        try:
            llvm_func = ctx.get_llvm_function(name)
            return llvm_func
        except Exception as e:
            pass


        # args: 1) ptr to input vector
        #       2) sizeof vector
        #       3) ptr to output vector
        float_ptr_ty = ctx.float_ty.as_pointer()
        args = [float_ptr_ty, ctx.int32_ty, float_ptr_ty]

        builder = ctx.create_llvm_function(args, self, name)
        llvm_func = builder.function
        llvm_func.attributes.add('alwaysinline')
        input_vector, dim, output_vector = llvm_func.args

        def get_fct_param_value(param_name):
            val = node.function.get_current_function_param(
                param_name, context)
            if val is None:
                val = node.function.get_current_function_param(
                    param_name, None)
            return ctx.float_ty(val[0])

        if isinstance(node.function, Linear):
            slope = get_fct_param_value('slope')
            intercept = get_fct_param_value('intercept')
            def modify_value(x):
                ret = builder.fadd(builder.fmul(x, slope), intercept)
                return ret

        elif isinstance(node.function, Logistic):
            neg_one = ctx.float_ty(-1)
            gain = builder.fmul(neg_one, get_fct_param_value('gain'))
            bias = get_fct_param_value('bias')
            offset = get_fct_param_value('offset')
            one = ctx.float_ty(1)
            exp = ctx.get_llvm_function("__pnl_builtin_exp")
            def modify_value(x):
                arg = builder.fsub(x, bias)
                arg = builder.fmul(gain, arg)
                arg = builder.fadd(arg, offset)

                ret = builder.call(exp, [arg])
                ret = builder.fadd(one, ret)
                ret = builder.fdiv(one, ret)
                return ret

        else:
            raise Exception(f"Unsupported compiled activation function {node.function}")

        #do computations
        with pnlvm.helpers.for_loop_zero_inc(builder, dim, "function_loop") as (b1, iterator):
            val_ptr = b1.gep(input_vector,[iterator])
            val = b1.load(val_ptr)
            val = modify_value(val)
            output_location = b1.gep(output_vector,[iterator])
            b1.store(val,output_location)

        builder.ret_void()

        return llvm_func

    # Helper method that creates a bin func that returns the derivative of the function into the builder
    # FIXME: Add compiled derivative functions, and move these calls there
    @handle_external_context()
    def bin_function_derivative_creator(self, ctx, node, context=None):
        # first try to get cached func
        name = node.name+"_"+node.function.name+"_derivative"
        try:
            llvm_func = ctx.get_llvm_function(name)
            return llvm_func
        except Exception as e:
            pass


        # args: 1) ptr to input vector
        #       2) sizeof vector
        #       3) ptr to output vector
        float_ptr_ty = ctx.float_ty.as_pointer()
        args = [float_ptr_ty, ctx.int32_ty, float_ptr_ty]
        builder = ctx.create_llvm_function(args, self,name )
        llvm_func = builder.function
        llvm_func.attributes.add('alwaysinline')

        input_vector, dim, output_vector = llvm_func.args
        def get_fct_param_value(param_name):
            val = node.function.get_current_function_param(
                param_name, context)
            if val is None:
                val = node.function.get_current_function_param(
                    param_name, None)
            return ctx.float_ty(val[0])

        if isinstance(node.function, Linear): # f(x) = mx + b, f'(x) = m
            slope = get_fct_param_value('slope')
            def modify_value(x):
                return slope

        elif isinstance(node.function, Logistic):# f'(x) = f(x)(1-f(x))

            neg_one = ctx.float_ty(-1)
            gain = builder.fmul(neg_one, get_fct_param_value('gain'))
            bias = get_fct_param_value('bias')
            offset = get_fct_param_value('offset')
            one = ctx.float_ty(1)
            exp = ctx.get_llvm_function("__pnl_builtin_exp")

            def modify_value(x):
                arg = builder.fsub(x, bias)
                arg = builder.fmul(gain, arg)
                arg = builder.fadd(arg, offset)

                f_x = builder.call(exp, [arg])
                f_x = builder.fadd(one, f_x)
                f_x = builder.fdiv(one, f_x)

                ret = builder.fsub(one ,f_x)
                ret = builder.fmul(f_x, ret)
                return ret

        else:
            raise Exception(f"Function type {node.function} is currently unsupported by compiled execution!")

        # do computations
        with pnlvm.helpers.for_loop_zero_inc(builder, dim, "derivative_loop") as (builder, iterator):
            val_ptr = builder.gep(input_vector,[iterator])
            val = builder.load(val_ptr)
            val = modify_value(val)
            output_location = builder.gep(output_vector,[iterator])
            builder.store(val,output_location)

        builder.ret_void()

        return llvm_func

    # helper method that identifies the type of function used by a node, gets the function
    # parameters and uses them to create a function object representing the function, then returns it
    def function_creator(self, node, context=None):
        def get_fct_param_value(param_name):
            val = node.function.get_current_function_param(
                param_name, context)
            if val is None:
                val = node.function.get_current_function_param(
                    param_name, Context(execution_id=None))
            return float(val)

        if isinstance(node.function, Linear):
            slope = get_fct_param_value('slope')
            intercept = get_fct_param_value('intercept')
            return lambda x: x * slope + intercept

        elif isinstance(node.function, Logistic):
            gain = get_fct_param_value('gain')
            bias = get_fct_param_value('bias')
            offset = get_fct_param_value('offset')
            return lambda x: 1 / (1 + torch.exp(-gain * (x + bias) + offset))

        # if we have relu function (the only other kind of function allowed by the autodiff composition)
        else:
            gain = get_fct_param_value('gain')
            bias = get_fct_param_value('bias')
            leak = get_fct_param_value('leak')
            return lambda x: (torch.max(input=(x - bias), other=torch.tensor([0], device=self.device).double()) * gain +
                              torch.min(input=(x - bias), other=torch.tensor([0], device=self.device).double()) * leak)

    # returns dict mapping psyneulink projections to corresponding pytorch weights. Pytorch weights are copied
    # over from tensors inside Pytorch's Parameter data type to numpy arrays (and thus copied to a different
    # memory location). This keeps the weights - and Pytorch in general - away from the user
    def get_weights_for_projections(self):
        weights_in_numpy = {}
        for projection, weights in self.projections_to_pytorch_weights.items():
            weights_in_numpy[projection] = weights.detach().cpu().numpy().copy()
        return weights_in_numpy

    # returns dict mapping psyneulink mechanisms to corresponding pytorch biases, the same way as the above function.
    # If composition is initialized with "param_init_from_PNL" set to true, then no biases are created in Pytorch,
    # and when called, this function returns an empty list.
    def get_biases_for_mechanisms(self):
        biases_in_numpy = {}
        for mechanism, biases in self.mechanisms_to_pytorch_biases.items():
            biases_in_numpy[mechanism] = biases.detach().cpu().numpy().copy()
        return biases_in_numpy



from . import inputstate
from . import modulatorysignals
from . import outputstate
from . import parameterstate
from . import state

from .inputstate import *
from .modulatorysignals import *
from .outputstate import *
from .parameterstate import *
from .state import *

__all__ = list(inputstate.__all__)
__all__.extend(modulatorysignals.__all__)
__all__.extend(outputstate.__all__)
__all__.extend(parameterstate.__all__)
__all__.extend(state.__all__)



import numpy as np
import psyneulink as pnl
import pytest


class TestMechanism:

    @pytest.mark.mechanism
    @pytest.mark.parametrize(
        'mechanism, default_variable, result_variable',
        [
            (pnl.TransferMechanism, [0], np.array([[0]])),
            (pnl.IntegratorMechanism, [0], np.array([[0]])),
        ]
    )
    def test_transfer_mech_instantiation(self, mechanism, default_variable, result_variable):
        T = mechanism(default_variable=default_variable)

        assert T.defaults.variable == result_variable
        assert T.defaults.value == result_variable

        assert T.function.defaults.variable == result_variable
        assert T.function.defaults.value == result_variable

        assert T.input_state.defaults.variable == result_variable[0]
        assert T.input_state.defaults.value == result_variable[0]

        assert T.input_state.function.defaults.variable == result_variable[0]
        assert T.input_state.function.defaults.value == result_variable[0]

    @pytest.mark.mechanism
    @pytest.mark.parametrize(
        'mechanism_type, default_variable, mechanism_value, function_value',
        [
            (pnl.ObjectiveMechanism, [0, 0, 0], np.array([[0, 0, 0]]), np.array([[0, 0, 0]]))
        ]
    )
    def test_value_shapes(self, mechanism_type, default_variable, mechanism_value, function_value):
        M = mechanism_type(default_variable=default_variable)

        assert M.defaults.value.shape == mechanism_value.shape
        assert M.function.defaults.value.shape == function_value.shape






# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.

# *********************************************  ComparatorMechanism ***************************************************

"""

Overview
--------

A ComparatorMechanism is a subclass of `ObjectiveMechanism` that receives two inputs (a sample and a target), compares
them using its `function <ComparatorMechanism.function>`, and places the calculated discrepancy between the two in its
*OUTCOME* `OutputState <ComparatorMechanism.output_state>`.

.. _ComparatorMechanism_Creation:

Creating a ComparatorMechanism
------------------------------

ComparatorMechanisms are generally created automatically when other PsyNeuLink components are created (such as
`LearningMechanisms <LearningMechanism_Creation>`).  A ComparatorMechanism can also be created directly by calling
its constructor.  Its **sample** and **target** arguments are used to specify the OutputStates that provide the
sample and target inputs, respectively (see `ObjectiveMechanism_Monitored_States` for details concerning their
specification, which are special versions of an ObjectiveMechanism's **monitor** argument).  When the
ComparatorMechanism is created, two InputStates are created, one each for its sample and target inputs (and named,
by default, *SAMPLE* and *TARGET*). Each is assigned a MappingProjection from the corresponding OutputState specified
in the **sample** and **target** arguments.

It is important to recognize that the value of the *SAMPLE* and *TARGET* InputStates must have the same length and type,
so that they can be compared using the ComparatorMechanism's `function <ComparatorMechanism.function>`.  By default,
they use the format of the OutputStates specified in the **sample** and **target** arguments, respectively,
and the `MappingProjection` to each uses an `IDENTITY_MATRIX`.  Therefore, for the default configuration, the
OutputStates specified in the **sample** and **target** arguments must have values of the same length and type.
If these differ, the **input_states** argument can be used to explicitly specify the format of the ComparatorMechanism's
*SAMPLE* and *TARGET* InputStates, to insure they are compatible with one another (as well as to customize their
names, if desired).  If the **input_states** argument is used, *both* the sample and target InputStates must be
specified.  Any of the formats for `specifying InputStates <InputState_Specification>` can be used in the argument.
If values are assigned for the InputStates, they must be of equal length and type.  Their types must
also be compatible with the value of the OutputStates specified in the **sample** and **target** arguments.  However,
the length specified for an InputState can differ from its corresponding OutputState;  in that case, by default, the
MappingProjection created uses a `FULL_CONNECTIVITY` matrix.  Thus, OutputStates of differing lengths can be mapped
to the sample and target InputStates of a ComparatorMechanism (see the `example <ComparatorMechanism_Example>` below),
so long as the latter are of the same length.  If a projection other than a `FULL_CONNECTIVITY` matrix is needed, this
can be specified using the *PROJECTION* entry of a `State specification dictionary <State_Specification>` for the
InputState in the **input_states** argument.

.. _ComparatorMechanism_Structure:

Structure
---------

A ComparatorMechanism has two `input_states <ComparatorMechanism.input_states>`, each of which receives a
`MappingProjection` from a corresponding OutputState specified in the **sample** and **target** arguments of its
constructor.  The InputStates are listed in the Mechanism's `input_states <ComparatorMechanism.input_States>` attribute
and named, respectively, *SAMPLE* and *TARGET*.  The OutputStates from which they receive their projections (specified
in the **sample** and **target** arguments) are listed in the Mechanism's `sample <ComparatorMechanism.sample>` and
`target <ComparatorMechanism.target>` attributes as well as in its `monitor <ComparatorMechanism.monitor>` attribute.
The ComparatorMechanism's `function <ComparatorMechanism.function>` compares the value of the sample and target
InputStates.  By default, it uses a `LinearCombination` function, assigning the sample InputState a `weight
<LinearCombination.weight>` of *-1* and the target a `weight <LinearCombination.weight>` of *1*, so that the sample
is subtracted from the target.  However, the `function <ComparatorMechanism.function>` can be customized, so long as
it is replaced with one that takes two arrays with the same format as its inputs and generates a similar array as its
result. The result is assigned as the value of the Comparator Mechanism's *OUTCOME* (`primary <OutputState_Primary>`)
OutputState.

.. _ComparatorMechanism_Function:

Execution
---------

When a ComparatorMechanism is executed, it updates its input_states with the values of the OutputStates specified
in its **sample** and **target** arguments, and then uses its `function <ComparatorMechanism.function>` to
compare these.  By default, the result is assigned to the `value <ComparatorMechanism.value>` of its *OUTCOME*
`output_state <ComparatorMechanism.output_state>`, and as the first item of the Mechanism's
`output_values <ComparatorMechanism.output_values>` attribute.

.. _ComparatorMechanism_Example:

Example
-------

.. _ComparatorMechanism_Default_Input_Value_Example:

*Formatting InputState values*

The **default_variable** argument can be used to specify a particular format for the SAMPLE and/or TARGET InputStates
of a ComparatorMechanism.  This can be useful when one or both of these differ from the format of the
OutputState(s) specified in the **sample** and **target** arguments. For example, for `Reinforcement Learning
<Reinforcement>`, a ComparatorMechanism is used to monitor an action selection Mechanism (the sample), and compare
this with a reinforcement signal (the target).  In the example below, the action selection Mechanism is a
`TransferMechanism` that uses the `SoftMax` function (and the `PROB <Softmax.PROB>` as its output format) to select
an action.  This generates a vector with a single non-zero value (the selected action). Because the output is a vector,
specifying it as the ComparatorMechanism's **sample** argument will generate a corresponding InputState with a vector
as its value.  This will not match the reward signal specified in the ComparatorMechanism's **target** argument, the
value of which is a single scalar.  This can be dealt with by explicitly specifying the format for the SAMPLE and
TARGET InputStates in the **default_variable** argument of the ComparatorMechanism's constructor, as follows::

    >>> import psyneulink as pnl
    >>> my_action_selection_mech = pnl.TransferMechanism(size=5,
    ...                                                  function=pnl.SoftMax(output=pnl.PROB))

    >>> my_reward_mech = pnl.TransferMechanism()

    >>> my_comparator_mech = pnl.ComparatorMechanism(default_variable = [[0],[0]],
    ...                                              sample=my_action_selection_mech,
    ...                                              target=my_reward_mech)

Note that ``my_action_selection_mechanism`` is specified to take an array of length 5 as its input, and therefore
generate one of the same length as its `primary output <OutputState_Primary>`.  Since it is assigned as the **sample**
of the ComparatorMechanism, by default this will create a *SAMPLE* InputState of length 5, that will not match the
length of the *TARGET* InputState (the default for which is length 1).  This is taken care of, by specifying the
**default_variable** argument as an array with two single-value arrays (corresponding to the *SAMPLE* and *TARGET*
InputStates). (In this example, the **sample** and **target** arguments are specified as Mechanisms since,
by default, each has only a single (`primary <OutputState_Primary>`) OutputState, that will be used;  if either had
more than one OutputState, and one of those was desired, it would have had to be specified explicitly in the
**sample** or **target** argument).

.. _ComparatorMechanism_Class_Reference:

Class Reference
---------------

"""

from collections.abc import Iterable

import numpy as np
import typecheck as tc

from psyneulink.core.components.functions.combinationfunctions import LinearCombination
from psyneulink.core.components.mechanisms.mechanism import Mechanism_Base
from psyneulink.core.components.mechanisms.processing.objectivemechanism import ObjectiveMechanism
from psyneulink.core.components.shellclasses import Mechanism
from psyneulink.core.components.states.inputstate import InputState
from psyneulink.core.components.states.outputstate import OutputState, PRIMARY, StandardOutputStates
from psyneulink.core.components.states.state import _parse_state_spec
from psyneulink.core.globals.context import Context, ContextFlags
from psyneulink.core.globals.keywords import COMPARATOR_MECHANISM, FUNCTION, INPUT_STATES, NAME, OUTCOME, SAMPLE, TARGET, VARIABLE, kwPreferenceSetName
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.componentpreferenceset import is_pref_set, kpReportOutputPref
from psyneulink.core.globals.preferences.preferenceset import PreferenceEntry, PreferenceLevel
from psyneulink.core.globals.utilities import is_numeric, is_value_spec, iscompatible, kwCompatibilityLength, kwCompatibilityNumeric, recursive_update
from psyneulink.core.globals.utilities import safe_len

__all__ = [
    'COMPARATOR_OUTPUT', 'ComparatorMechanism', 'ComparatorMechanismError', 'MSE', 'SSE',
]

SSE = 'SSE'
MSE = 'MSE'


class COMPARATOR_OUTPUT():
    """
    .. _ComparatorMechanism_Standard_OutputStates:

    `Standard OutputStates <OutputState_Standard>` for `ComparatorMechanism`

    .. _COMPARATOR_MECHANISM_SSE

    *SSE*
        the value of the sum squared error of the Mechanism's function

    .. _COMPARATOR_MECHANISM_MSE

    *MSE*
        the value of the mean squared error of the Mechanism's function

    """
    SSE = SSE
    MSE = MSE


class ComparatorMechanismError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)


class ComparatorMechanism(ObjectiveMechanism):
    """
    ComparatorMechanism(                                \
        sample,                                         \
        target,                                         \
        input_states=[SAMPLE,TARGET]                    \
        function=LinearCombination(weights=[[-1],[1]],  \
        output_states=OUTCOME                           \
        params=None,                                    \
        name=None,                                      \
        prefs=None)

    Subclass of `ObjectiveMechanism` that compares the values of two `OutputStates <OutputState>`.

    COMMENT:
        Description:
            ComparatorMechanism is a subtype of the ObjectiveMechanism Subtype of the ProcssingMechanism Type
            of the Mechanism Category of the Component class.
            By default, it's function uses the LinearCombination Function to compare two input variables.
            COMPARISON_OPERATION (functionParams) determines whether the comparison is subtractive or divisive
            The function returns an array with the Hadamard (element-wise) differece/quotient of target vs. sample,
                as well as the mean, sum, sum of squares, and mean sum of squares of the comparison array

        Class attributes:
            + componentType (str): ComparatorMechanism
            + classPreference (PreferenceSet): Comparator_PreferenceSet, instantiated in __init__()
            + classPreferenceLevel (PreferenceLevel): PreferenceLevel.SUBTYPE
            + class_defaults.variable (value):  Comparator_DEFAULT_STARTING_POINT // QUESTION: What to change here
            + paramClassDefaults (dict): {FUNCTION_PARAMS:{COMPARISON_OPERATION: SUBTRACTION}}

        Class methods:
            None

        MechanismRegistry:
            All instances of ComparatorMechanism are registered in MechanismRegistry, which maintains an
              entry for the subclass, a count for all instances of it, and a dictionary of those instances
    COMMENT

    Arguments
    ---------

    sample : OutputState, Mechanism, value, or string
        specifies the value to compare with the `target` by the `function <ComparatorMechanism.function>`.

    target :  OutputState, Mechanism, value, or string
        specifies the value with which the `sample` is compared by the `function <ComparatorMechanism.function>`.

    input_states :  List[InputState, value, str or dict] or Dict[] : default [SAMPLE, TARGET]
        specifies the names and/or formats to use for the values of the sample and target InputStates;
        by default they are named *SAMPLE* and *TARGET*, and their formats are match the value of the OutputStates
        specified in the **sample** and **target** arguments, respectively (see `ComparatorMechanism_Structure`
        for additional details).

    function :  Function, function or method : default Distance(metric=DIFFERENCE)
        specifies the `function <Comparator.function>` used to compare the `sample` with the `target`.

    output_states :  List[OutputState, value, str or dict] or Dict[] : default [OUTCOME]
        specifies the OutputStates for the Mechanism;

    params :  Optional[Dict[param keyword: param value]]
        a `parameter dictionary <ParameterState_Specification>` that can be used to specify the parameters for
        the Mechanism, its function, and/or a custom function and its parameters. Values specified for parameters in
        the dictionary override any assigned to those parameters in arguments of the
        constructor.

    name : str : default see `name <ComparatorMechanism.name>`
        specifies the name of the ComparatorMechanism.

    prefs : PreferenceSet or specification dict : default Mechanism.classPreferences
        specifies the `PreferenceSet` for the ComparatorMechanism; see `prefs <ComparatorMechanism.prefs>` for details.


    Attributes
    ----------

    COMMENT:
    default_variable : Optional[List[array] or 2d np.array]
    COMMENT

    sample : OutputState
        determines the value to compare with the `target` by the `function <ComparatorMechanism.function>`.

    target : OutputState
        determines the value with which `sample` is compared by the `function <ComparatorMechanism.function>`.

    input_states : ContentAddressableList[InputState, InputState]
        contains the two InputStates named, by default, *SAMPLE* and *TARGET*, each of which receives a
        `MappingProjection` from the OutputStates referenced by the `sample` and `target` attributes
        (see `ComparatorMechanism_Structure` for additional details).

    function : CombinationFunction, function or method
        used to compare the `sample` with the `target`.  It can be any PsyNeuLink `CombinationFunction`,
        or a python function that takes a 2d array with two items and returns a 1d array of the same length
        as the two input items.

    value : 1d np.array
        the result of the comparison carried out by the `function <ComparatorMechanism.function>`.

    output_state : OutputState
        contains the `primary <OutputState_Primary>` OutputState of the ComparatorMechanism; the default is
        its *OUTCOME* OutputState, the value of which is equal to the `value <ComparatorMechanism.value>`
        attribute of the ComparatorMechanism.

    output_states : ContentAddressableList[OutputState]
        contains, by default, only the *OUTCOME* (primary) OutputState of the ComparatorMechanism.

    output_values : 2d np.array
        contains one item that is the value of the *OUTCOME* OutputState.

    name : str
        the name of the ComparatorMechanism; if it is not specified in the **name** argument of the constructor, a
        default is assigned by MechanismRegistry (see `Naming` for conventions used for default and duplicate names).

    prefs : PreferenceSet or specification dict
        the `PreferenceSet` for the ComparatorMechanism; if it is not specified in the **prefs** argument of the
        constructor, a default is assigned using `classPreferences` defined in __init__.py (see :doc:`PreferenceSet
        <LINK>` for details).


    """
    componentType = COMPARATOR_MECHANISM

    classPreferenceLevel = PreferenceLevel.SUBTYPE
    # These will override those specified in TypeDefaultPreferences
    classPreferences = {
        kwPreferenceSetName: 'ComparatorCustomClassPreferences',
        kpReportOutputPref: PreferenceEntry(False, PreferenceLevel.INSTANCE)}

    class Parameters(ObjectiveMechanism.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <ComparatorMechanism.variable>`

                    :default value: numpy.array([[0], [0]])
                    :type: numpy.ndarray
                    :read only: True

                function
                    see `function <ComparatorMechanism.function>`

                    :default value: `LinearCombination`(weights=numpy.array([[-1], [ 1]]))
                    :type: `Function`

                sample
                    see `sample <ComparatorMechanism.sample>`

                    :default value: None
                    :type:

                target
                    see `target <ComparatorMechanism.target>`

                    :default value: None
                    :type:

        """
        # By default, ComparatorMechanism compares two 1D np.array input_states
        variable = Parameter(np.array([[0], [0]]), read_only=True)
        function = Parameter(LinearCombination(weights=[[-1], [1]]), stateful=False, loggable=False)
        sample = None
        target = None

    # ComparatorMechanism parameter and control signal assignments):
    paramClassDefaults = Mechanism_Base.paramClassDefaults.copy()

    standard_output_states = ObjectiveMechanism.standard_output_states.copy()

    standard_output_states.extend([{NAME: SSE,
                                    FUNCTION: lambda x: np.sum(x*x)},
                                   {NAME: MSE,
                                    FUNCTION: lambda x: np.sum(x * x) / safe_len(x)}])

    @tc.typecheck
    def __init__(self,
                 default_variable=None,
                 sample: tc.optional(tc.any(OutputState, Mechanism_Base, dict, is_numeric, str))=None,
                 target: tc.optional(tc.any(OutputState, Mechanism_Base, dict, is_numeric, str))=None,
                 function=LinearCombination(weights=[[-1], [1]]),
                 output_states:tc.optional(tc.any(str, Iterable))=(OUTCOME,),
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 **kwargs
                 ):

        input_states = kwargs.pop(INPUT_STATES, {})
        if input_states:
            input_states = {INPUT_STATES: input_states}

        input_states = self._merge_legacy_constructor_args(sample, target, default_variable, input_states)

        # Default output_states is specified in constructor as a tuple rather than a list
        # to avoid "gotcha" associated with mutable default arguments
        # (see: bit.ly/2uID3s3 and http://docs.python-guide.org/en/latest/writing/gotchas/)
        if isinstance(output_states, (str, tuple)):
            output_states = list(output_states)

        # IMPLEMENTATION NOTE: The following prevents the default from being updated by subsequent assignment
        #                     (in this case, to [OUTCOME, {NAME= MSE}]), but fails to expose default in IDE
        # output_states = output_states or [OUTCOME, MSE]

        # Create a StandardOutputStates object from the list of stand_output_states specified for the class
        if not isinstance(self.standard_output_states, StandardOutputStates):
            self.standard_output_states = StandardOutputStates(self,
                                                               self.standard_output_states,
                                                               indices=PRIMARY)

        super().__init__(monitor=input_states,
                         function=function,
                         output_states=output_states.copy(), # prevent default from getting overwritten by later assign
                         params=params,
                         name=name,
                         prefs=prefs,

                         **kwargs
                         )

        # Require Projection to TARGET InputState (already required for SAMPLE as primary InputState)
        self.input_states[1].parameters.require_projection_in_composition._set(True, Context())

    def _validate_params(self, request_set, target_set=None, context=None):
        """If sample and target values are specified, validate that they are compatible
        """

        if INPUT_STATES in request_set and request_set[INPUT_STATES] is not None:
            input_states = request_set[INPUT_STATES]

            # Validate that there are exactly two input_states (for sample and target)
            num_input_states = len(input_states)
            if num_input_states != 2:
                raise ComparatorMechanismError("{} arg is specified for {} ({}), so it must have exactly 2 items, "
                                               "one each for {} and {}".
                                               format(INPUT_STATES,
                                                      self.__class__.__name__,
                                                      len(input_states),
                                                      SAMPLE,
                                                      TARGET))

            # Validate that input_states are specified as dicts
            if not all(isinstance(input_state,dict) for input_state in input_states):
                raise ComparatorMechanismError("PROGRAM ERROR: all items in input_state args must be converted to dicts"
                                               " by calling State._parse_state_spec() before calling super().__init__")

            # Validate length of variable for sample = target
            if VARIABLE in input_states[0]:
                # input_states arg specified in standard state specification dict format
                lengths = [len(input_state[VARIABLE]) for input_state in input_states]
            else:
                # input_states arg specified in {<STATE_NAME>:<STATE SPECIFICATION DICT>} format
                lengths = [len(list(input_state_dict.values())[0][VARIABLE]) for input_state_dict in input_states]

            if lengths[0] != lengths[1]:
                raise ComparatorMechanismError("Length of value specified for {} InputState of {} ({}) must be "
                                               "same as length of value specified for {} ({})".
                                               format(SAMPLE,
                                                      self.__class__.__name__,
                                                      lengths[0],
                                                      TARGET,
                                                      lengths[1]))

        elif SAMPLE in request_set and TARGET in request_set:

            sample = request_set[SAMPLE]
            if isinstance(sample, InputState):
                sample_value = sample.value
            elif isinstance(sample, Mechanism):
                sample_value = sample.input_value[0]
            elif is_value_spec(sample):
                sample_value = sample
            else:
                sample_value = None

            target = request_set[TARGET]
            if isinstance(target, InputState):
                target_value = target.value
            elif isinstance(target, Mechanism):
                target_value = target.input_value[0]
            elif is_value_spec(target):
                target_value = target
            else:
                target_value = None

            if sample is not None and target is not None:
                if not iscompatible(sample, target, **{kwCompatibilityLength: True,
                                                       kwCompatibilityNumeric: True}):
                    raise ComparatorMechanismError("The length of the sample ({}) must be the same as for the target ({})"
                                                   "for {} {}".
                                                   format(len(sample),
                                                          len(target),
                                                          self.__class__.__name__,
                                                          self.name))

        super()._validate_params(request_set=request_set,
                                 target_set=target_set,
                                 context=context)

    def _merge_legacy_constructor_args(self, sample, target, default_variable=None, input_states=None):

        # USE sample and target TO CREATE AN InputState specfication dictionary for each;
        # DO SAME FOR InputStates argument, USE TO OVERWRITE ANY SPECIFICATIONS IN sample AND target DICTS
        # TRY tuple format AS WAY OF PROVIDED CONSOLIDATED variable AND OutputState specifications

        sample_dict = _parse_state_spec(owner=self,
                                        state_type=InputState,
                                        state_spec=sample,
                                        name=SAMPLE)

        target_dict = _parse_state_spec(owner=self,
                                        state_type=InputState,
                                        state_spec=target,
                                        name=TARGET)

        # If either the default_variable arg or the input_states arg is provided:
        #    - validate that there are exactly two items in default_variable or input_states list
        #    - if there is an input_states list, parse it and use it to update sample and target dicts
        if input_states:
            input_states = input_states[INPUT_STATES]
            # print("type input_states = {}".format(type(input_states)))
            if not isinstance(input_states, list):
                raise ComparatorMechanismError("If an \'{}\' argument is included in the constructor for a {} "
                                               "it must be a list with two {} specifications.".
                                               format(INPUT_STATES, ComparatorMechanism.__name__, InputState.__name__))

        input_states = input_states or default_variable

        if input_states is not None:
            if len(input_states)!=2:
                raise ComparatorMechanismError("If an \'input_states\' arg is "
                                               "included in the constructor for "
                                               "a {}, it must be a list with "
                                               "exactly two items (not {})".
                                               format(ComparatorMechanism.__name__, len(input_states)))

            sample_input_state_dict = _parse_state_spec(owner=self,
                                                        state_type=InputState,
                                                        state_spec=input_states[0],
                                                        name=SAMPLE,
                                                        value=None)

            target_input_state_dict = _parse_state_spec(owner=self,
                                                        state_type=InputState,
                                                        state_spec=input_states[1],
                                                        name=TARGET,
                                                        value=None)

            sample_dict = recursive_update(sample_dict, sample_input_state_dict)
            target_dict = recursive_update(target_dict, target_input_state_dict)

        return [sample_dict, target_dict]



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ******************************************  GatingSignal *****************************************************

"""
Overview
--------

A GatingSignal is a type of `ModulatorySignal <ModulatorySignal>` that is specialized for use with a `GatingMechanism`
and one or more `GatingProjections <GatingProjection>`, to modify the `value <State_Base.value>`\\(s) of the
`InputState(s) <InputState>` and/or `OutputState(s) <OutputState>` to which they project. A GatingSignal receives the
value from the `gating_allocation <GatingMechanism.gating_allocation>` of the GatingMechanism to which it belongs,
and assigns that as the value of its `gating_signal <GatingSignal.gating_signal>` to its `GatingProjection(s)
<GatingProjection>`, each of which projects to an InputState or OutputState and is used to modulate the `value
<State_Base.value>` of that State.


.. _GatingSignal_Creation:

Creating a GatingSignal
-----------------------

A GatingSignal is created automatically whenever an `InputState` or `OutputState` of a `Mechanism <Mechanism>` is
specified for gating.  This can be done either in the **gating_signals** argument of the constructor for a
`GatingMechanism <GatingMechanism_GatingSignals>`, or in the `specification of projections <State_Projections>` to
the InputState or OutputState.  Although a GatingSignal can be created directly using its constructor (or any of the
other ways for `creating an OutputState <OutputStates_Creation>`), this is usually not necessary nor is it advisable,
as a GatingSignal has dedicated components and requirements for configuration that must be met for it to function
properly.

.. _GatingSignal_Specification:

*Specifying GatingSignals*
~~~~~~~~~~~~~~~~~~~~~~~~~~

When a GatingSignal is specified in the **gating_signals** argument of the constructor for a `GatingMechanism`, the
InputState(s) and/or OutputState(s) it gates must be specified. This can take any of the following forms:

  * **InputState** or **OutputState** of a Mechanism;
  ..
  * **Mechanism** -- the `primary `InputState <InputState_Primary>` or `OutputState <OutputState_Primary>` is used;
  ..
  * **specification dictionary** -- can take either of the following two forms:

    * for gating a single state, the dictionary can have the following two entries:

        * *NAME*: str
            the string must be the name of the State to be gated; the GatingSignal will named by appending
            "_GatingSignal" to the name of the State.

        * *MECHANISM*: Mechanism
            the Mechanism must be the one to the which the State to be gated belongs.

    * for gating multiple states, the dictionary can have the following entry:

        * <str>:list
            the string used as the key specifies the name to be used for the GatingSignal,
            and each item of the list must be a `specification of a State <State_Creation>` to be
            gated by the GatingSignal (and that will receive a `GatingProjection` from it).

    The dictionary can also contain entries for any other GatingSignal attributes to be specified
    (e.g., a *MODULATION* entry, the value of which determines how the GatingSignal modulates the
    `value <State_Base.value>` of the State(s) that it gates; or a *VARIABLE* entry specifying which item
    of the GatingMechanism's `gating_allocation <GatingMechanism.gating_allocation>` it should use as its `value
    <GatingSignal,value>`;  see `OutputState_Customization`).
  ..
  * **2-item tuple:** *(<State name or list of State names>, <Mechanism>)* -- the 1st item must be the name of the
    State (or list of State names), and the 2nd item the Mechanism to which it (they) belong(s); this is a convenience
    format, which is simpler to use than a specification dictionary (see below), but precludes specification of
    `parameters <GatingSignal_Structure>` for the GatingSignal.

.. _GatingSignal_Structure:

Structure
---------

A GatingSignal is owned by a `GatingMechanism`, and associated with one or more `GatingProjections <GatingProjection>`,
each of which projects to the InputState or OutputState that it gates.

.. _GatingSignal_Projections:

*Projections*
~~~~~~~~~~~~~

When a GatingSignal is created, it can be assigned one or more `GatingProjections <GatingProjection>`, using either
the **projections** argument of its constructor, or in an entry of a dictionary assigned to the **params** argument
with the key *PROJECTIONS*.  These will be assigned to its `efferents  <GatingSignal.efferents>` attribute.  See
`State Projections <State_Projections>` for additional details concerning the specification of Projections when
creating a State.

.. note::
   Although a GatingSignal can be assigned more than one `GatingProjection`, all of those Projections will convey
   the same `gating_signal <LearningMechanism>` (received from the GatingMechanism to which the GatingSignal
   belongs), and use the same form of `modulation <GatingSignal_Modulation>`.  This can be useful for implementing
   broadly projecting modulatory effects.

.. _GatingSignal_Modulation:

*Modulation*
~~~~~~~~~~~~

Each GatingSignal has a `modulation <GatingSignal.modulation>` attribute that determines how the GatingSignal's
`value <GatingSignal.value>` (i.e., its `gating_signal <GatingSignal.gating_signal>`) is used by the States to which it
projects to modify their `value <State_Base.value>` \\s (see `ModulatorySignal_Modulation` for an explanation of how the
`modulation <GatingSignal.modulation>` attribute is specified and used to modulate the `value <State_Base.value>` of a
State). The `modulation <GatingSignal.modulation>` attribute can be specified in the **modulation** argument of the
constructor for a GatingSignal, or in a specification dictionary as described `above <GatingSignal_Specification>`.
The value must be a value of `ModulationParam`;  if it is not specified, its default is the value of the `modulation
<GatingMechanism.modulation>` attribute of the GatingMechanism to which the GatingSignal belongs (which is the same
for all of the GatingSignals belonging to that GatingMechanism).  The value of the
`modulation <GatingSignal.modulation>` attribute of a GatingSignal is used by all of the
`GatingProjections <GatingProjection>` that project from that GatingSignal.


.. _GatingSignal_Execution:

Execution
---------

A GatingSignal cannot be executed directly.  It is executed whenever the `GatingMechanism` to which it belongs is
executed.  When this occurs, the GatingMechanism provides the GatingSignal with one of the values from its
`gating_allocation <GatingMechanism.gating_signal>`, that is used by its `function <GatingSignal.function>` to generate its
the value of its `gating_signal <GatingSignal.gating_signal>`.  That, in turn, is used by its `GatingProjection(s)
<GatingProjection>` to modulate the `value <State_Base.value>` of the States to which they project. How the modulation
is executed is determined by the GatingSignal's `modulation <GatingSignal.modulation>` attribute
(see `above <GatingSigna_Modulation>`, and `ModulatorySignal_Modulation` for a more detailed explanation of how
modulation operates).

.. note::
   The change in the `value <State_Base.value>` of InputStates and OutputStates in response to the execution of a
   GatingSignal are not applied until the Mechanism(s) to which those states belong are next executed;
   see :ref:`Lazy Evaluation <LINK>` for an explanation of "lazy" updating).

.. _GatingSignal_Examples:

Examples
--------

**Gate an InputState and OutputState**.  In the following example, `GatingMechanism` is configured to gate the
`primary InputState <InputState_Primary>` of one Mechanism, and the `primary OutputState <OutputState_Primary>`
of another::

    >>> import psyneulink as pnl
    >>> my_mechanism_A = pnl.TransferMechanism(name="Mechanism A")
    >>> my_mechanism_B = pnl.TransferMechanism(name="Mechanism B")
    >>> my_gating_mechanism = pnl.GatingMechanism(gating_signals=[my_mechanism_A.input_state,
    ...                                                           my_mechanism_B.output_state])

Note that, in the **gating_signals** argument, the first item references a Mechanism (``my_mechanism_A``) rather than
one of its states -- this is all that is necessary, since the default for a `GatingSignal` is to modulate the
`primary InputState <InputState_Primary>` of a Mechanism.  The second item explicitly specifies the State to be gated,
since it is an OutputState.  This will generate two GatingSignals, each of which will multiplicatively modulate the
value of the State to which it projects.  This is because, by default, the `modulation <GatingSignal.modulation>`
attribute of a GatingSignal is the *MULTIPLICATIVE_PARAM* for the `function <State_Base.function>` of the State to which
it projects.  For an InputState, the default `function <InputState.function>` is `Linear` and its *MULTIPLICATIVE_PARAM*
is its `slope <Linear.slope>` parameter.  Thus, the value of the GatingSignal is assigned to the slope, which multiplies
the State`s `variable <State_Base.variable>` (i.e., its input(s)) to determine its `value <State_Base.value>`.

**Modulate the InputStates of several Mechanisms**.  In next example, a `GatingMechanism` is created that modulates
the `InputState` of all the layers in a 3-layered feedforward neural network.  Ordinarily, gating modulates the
*MULTIPLICATIVE_PARAM* of an InputState's `function <InputState.function>`.  In the example, this is changed so that
it *adds* the `value <GatingSignal.value>` of the `GatingSignal` to the `value <InputState.value>` of each InputState::

    >>> my_input_layer = pnl.TransferMechanism(size=3)
    >>> my_hidden_layer = pnl.TransferMechanism(size=5)
    >>> my_output_layer = pnl.TransferMechanism(size=2)
    >>> my_gating_mechanism = pnl.GatingMechanism(gating_signals=[{pnl.NAME: 'GATE_ALL',
    ...                                                            pnl.PROJECTIONS: [my_input_layer,
    ...                                                                              my_hidden_layer,
    ...                                                                              my_output_layer]}],
    ...                                           modulation=pnl.ADDITIVE)

Note that, again, the **gating_signals** are listed as Mechanisms, since in this case it is their primary InputStates
that are to be gated. Since they are all listed in a single entry of a
`specification dictionary <GatingSignal_Specification>`, they will all be gated by a single GatingSignal named
``GATE_ALL``, that will send a `GatingProjection` to the InputState of each of the Mechanisms listed (the next example
shows how different InputStates can be differentially gated by a `GatingMechanism`). Finally, note that the
`ModulationParam` specified for the `GatingMechanism` (and therefore the default for its GatingSignals) pertains to
the `function <InputState.function>` of each `InputState`. By default that is a `Linear` function, the *ADDITIVE_PARAM*
of which is its `intercept <Linear.intercept>` parameter. Therefore, in the example above, each time the InputStates
are updated, the value of the GatingSignal will be assigned as the `intercept` of each InputState's
`function <InputState.function>`, thus adding that amount to the input to the State before determining its
`value <InputState.value>`.

**Gate InputStates differentially**.  In the example above, the InputStates for all of the Mechanisms were gated
using a single GatingSignal.  In the example below, a different GatingSignal is assigned to the InputState of each
Mechanism::

    >>> my_gating_mechanism = pnl.GatingMechanism(gating_signals=[{pnl.NAME: 'GATING_SIGNAL_A',
    ...                                                            pnl.MODULATION: pnl.ADDITIVE,
    ...                                                            pnl.PROJECTIONS: my_input_layer},
    ...                                                           {pnl.NAME: 'GATING_SIGNAL_B',
    ...                                                            pnl.PROJECTIONS: [my_hidden_layer,
    ...                                                                              my_output_layer]}])

Here, two GatingSignals are specified as `specification dictionaries <GatingSignal_Specification>`, each of which
contains an entry for the name of the GatingSignal, and a *PROJECTIONS* entry that specifies the States to which the
GatingSignal should project (i.e., the ones to be gated).  Once again, the specifications exploit the fact that the
default is to gate the `primary InputState <InputState_Primary>` of a Mechanism, so those are what are referenced. The
first dict also contains a  *MODULATION* entry that specifies the value of the `modulation <GatingSignal.modulation>`
attribute for the GatingSignal.  The second one does not, so the default will be used (which, for a GatingSignal, is
`ModulationParam.MULTIPLICATIVE`).  Thus, the InputState of ``my_input_layer`` will be additively modulated by
``GATING_SIGNAL_A``, while the InputStates of ``my_hidden_layer`` and ``my_output_layer`` will be multiplicatively
modulated by ``GATING_SIGNAL_B``.

**Creating and assigning stand-alone GatingSignals**.  GatingSignals can also be created on their own, and then later
assigned to a GatingMechanism.  In the example below, the same GatingSignals specified in the previous example are
created directly and then assigned to ``my_gating_mechanism``::

    >>> my_gating_signal_A = pnl.GatingSignal(name='GATING_SIGNAL_A',
    ...                                       modulation=pnl.ADDITIVE,
    ...                                       projections=my_input_layer)
    >>> my_gating_signal_B = pnl.GatingSignal(name='GATING_SIGNAL_B',
    ...                                       projections=[my_hidden_layer,
    ...                                                    my_output_layer])
    >>> my_gating_mechanism = pnl.GatingMechanism(gating_signals=[my_gating_signal_A,
    ...                                                           my_gating_signal_B])


Class Reference
---------------

"""

import numpy as np
import typecheck as tc

from psyneulink.core.components.functions.transferfunctions import Linear
from psyneulink.core.components.states.modulatorysignals.modulatorysignal import ModulatorySignal, modulatory_signal_keywords
from psyneulink.core.components.states.outputstate import PRIMARY, SEQUENTIAL, _output_state_variable_getter
from psyneulink.core.components.states.state import State_Base
from psyneulink.core.globals.context import ContextFlags
from psyneulink.core.globals.defaults import defaultGatingAllocation
from psyneulink.core.globals.keywords import \
    CONTEXT, GATE, GATING_PROJECTION, GATING_SIGNAL, INPUT_STATE, INPUT_STATES, \
    OUTPUT_STATE, OUTPUT_STATES, OUTPUT_STATE_PARAMS, PROJECTIONS, PROJECTION_TYPE, RECEIVER, VARIABLE
from psyneulink.core.globals.parameters import Parameter
from psyneulink.core.globals.preferences.componentpreferenceset import is_pref_set
from psyneulink.core.globals.preferences.preferenceset import PreferenceLevel

__all__ = [
    'gating_signal_keywords', 'GatingSignal', 'GatingSignalError',
]


class GatingSignalError(Exception):
    def __init__(self, error_value):
        self.error_value = error_value

    def __str__(self):
        return repr(self.error_value)

gating_signal_keywords = {GATE}
gating_signal_keywords.update(modulatory_signal_keywords)


class GatingSignal(ModulatorySignal):
    """
    GatingSignal(                                   \
        owner,                                      \
        default_allocation=defaultGatingAllocation, \
        index=PRIMARY                               \
        variable=defaultGatingAllocation            \
        function=Linear(),                          \
        modulation=MULTIPLICATIVE,                  \
        projections=None,                           \
        params=None,                                \
        name=None,                                  \
        prefs=None)

    A subclass of `ModulatorySignal <ModulatorySignal>` used by a `GatingMechanism` to modulate the value(s)
    of one more `InputState(s) <InputState>` and/or `OutputState(s) <OutputState>`.

    COMMENT:

        Description
        -----------
            The GatingSignal class is a subtype of the OutputState class in the State category of Component,
            It is used primarily as the sender for GatingProjections
            Its FUNCTION updates its value:
                note:  currently, this is the identity function, that simply maps variable to self.value

        Class attributes:
            + componentType (str) = GATING_SIGNAL
            + paramClassDefaults (dict)
                + FUNCTION (Linear)
                + FUNCTION_PARAMS (Modulation.MULTIPLY)

        Class methods:
            function (executes function specified in params[FUNCTION];  default: Linear

        StateRegistry
        -------------
            All OutputStates are registered in StateRegistry, which maintains an entry for the subclass,
              a count for all instances of it, and a dictionary of those instances
    COMMENT


    Arguments
    ---------

    owner : GatingMechanism
        specifies the `GatingMechanism` to which to assign the GatingSignal.

    default_allocation : scalar, list or np.ndarray : defaultGatingAllocation
        specifies the template and default value used for `allocation <GatingSignal.allocation>`;  must match the
        shape of each item specified in `allocation_samples <GatingSignal.allocation_samples>`.

    index : int : default PRIMARY
        specifies the item of the owner GatingMechanism's `gating_allocation <GatingMechanism.gating_allocation>` used as the
        GatingSignal's `value <GatingSignal.value>`.

    function : Function or method : default Linear
        specifies the function used to determine the value of the GatingSignal from the value of its
        `owner <GatingMechanism.owner>`.

    modulation : ModulationParam : default ModulationParam.MULTIPLICATIVE
        specifies the way in which the `value <GatingSignal.value>` the GatingSignal is used to modify the
        `value <State_Base.value>` of the State(s) to which the GatingSignal's
        `GatingProjection(s) <GatingProjection>` project.

    modulates : list of Projection specifications
        specifies the `GatingProjection(s) <GatingProjection>` to be assigned to the GatingSignal, and that will be
        listed in its `efferents <GatingSignal.efferents>` attribute (see `GatingSignal_Projections` for additional
        details).

    params : Dict[param keyword: param value] : default None
        a `parameter dictionary <ParameterState_Specification>` that can be used to specify the parameters for
        the ControlSignal and/or a custom function and its parameters. Values specified for parameters in the dictionary
        override any assigned to those parameters in arguments of the constructor.

    name : str : default see ModulatorySignal `name <ModulatorySignal.name>`
        specifies the name of the GatingSignal;  see GatingSignal `name <ModulatorySignal.name>` for additional
        details.

    prefs : PreferenceSet or specification dict : default State.classPreferences
        specifies the `PreferenceSet` for the GatingSignal; see `prefs <GatingSignal.prefs>` for details.


    Attributes
    ----------

    owner : GatingMechanism
        the `GatingMechanism` to which the GatingSignal belongs.

    variable : scalar, list or np.ndarray
        same as `allocation <GatingSignal.allocation>`.

    allocation : float : default: defaultGatingAllocation
        value assigned by the GatingSignal's `owner <GatingSignal.owner>`, and used as the `variable
        <GatingSignal.variable>` of the it `function <GatingSignal.function>` to determine the GatingSignal's
        `GatingSignal.intensity`.
    COMMENT:
    FOR DEVELOPERS:  Implemented as an alias of the GatingSignal's variable Parameter
    COMMENT

    function : TransferFunction :  default Linear(slope=1, intercept=0)
        provides the GatingSignal's `value <GatingMechanism.value>`; the default is an identity function that
        passes the input to the GatingMechanism as value for the GatingSignal.

    value : number, list or np.ndarray
        result of the GatingSignal's `function <GatingSignal.function>`
        (same as its `gating_signal <GatingSignal.gating_signal>`).

    intensity : float
        result of the GatingSignal's `function <GatingSignal.function>`;
        assigned as the value of the GatingSignal's GatingProjection, and used to modify the `value <State_Base.value>`
        of the State(s) to which the GatingSignal's `GatingProjection(s) <GatingProjection>` project; same as
        `gating_signal <GatingSignal.gating_signal>`.

    index : int
        the item of the owner GatingMechanism's `gating_allocation <GatingMechanism.gating_allocation>` used as the
        GatingSignal's `value <GatingSignal.value>`.

    gating_signal : number, list or np.ndarray
        result of the GatingSignal's `function <GatingSignal.function>` (same as its `value <GatingSignal.value>`).


    modulation : ModulationParam
        determines the way in the which `value <GatingSignal.value>` of the GatingSignal is used to modify the
        `value <State_Base.value>` of the State(s) to which the GatingSignal's
        `GatingProjection(s) <GatingProjection>` project.

    efferents : [List[GatingProjection]]
        a list of the `GatingProjections <GatingProjection>` assigned to (i.e., that project from) the GatingSignal.

    name : str
        name of the GatingSignal; if not is specified in the **name** argument of its constructor, a default name
        is assigned (see `name <ModulatorySignal.name>`).

        .. note::
            Unlike other PsyNeuLink components, State names are "scoped" within a Mechanism, meaning that States with
            the same name are permitted in different Mechanisms.  However, they are *not* permitted in the same
            Mechanism: States within a Mechanism with the same base name are appended an index in the order of their
            creation.

    prefs : PreferenceSet or specification dict
        the `PreferenceSet` for the GatingSignal; if it is not specified in the **prefs** argument of the constructor,
        a default is assigned using `classPreferences` defined in __init__.py (see :doc:`PreferenceSet <LINK>` for
        details).

    """

    #region CLASS ATTRIBUTES

    componentType = GATING_SIGNAL
    componentName = 'GatingSignal'
    paramsType = OUTPUT_STATE_PARAMS

    stateAttributes = ModulatorySignal.stateAttributes | {GATE}

    connectsWith = [INPUT_STATE, OUTPUT_STATE]
    connectsWithAttribute = [INPUT_STATES, OUTPUT_STATES]
    projectionSocket = RECEIVER
    modulators = []

    classPreferenceLevel = PreferenceLevel.TYPE
    # Any preferences specified below will override those specified in TypeDefaultPreferences
    # Note: only need to specify setting;  level will be assigned to TYPE automatically
    # classPreferences = {
    #     kwPreferenceSetName: 'OutputStateCustomClassPreferences',
    #     kp<pref>: <setting>...}

    class Parameters(ModulatorySignal.Parameters):
        """
            Attributes
            ----------

                variable
                    see `variable <GatingSignal.variable>`

                    :default value: numpy.array([0.5])
                    :type: numpy.ndarray

                value
                    see `value <GatingSignal.value>`

                    :default value: numpy.array([0.5])
                    :type: numpy.ndarray
                    :read only: True

                allocation_samples
                    see `allocation_samples <GatingSignal.allocation_samples>`

                    :default value: None
                    :type:

        """
        variable = Parameter(np.array([defaultGatingAllocation]),
                             aliases='allocation',
                             getter=_output_state_variable_getter)
        value = Parameter(np.array([defaultGatingAllocation]), read_only=True, aliases=['intensity'])
        allocation_samples = Parameter(None, modulable=True)

    paramClassDefaults = State_Base.paramClassDefaults.copy()
    paramClassDefaults.update({
        PROJECTION_TYPE: GATING_PROJECTION,
        GATE:None,
    })
    #endregion

    @tc.typecheck
    def __init__(self,
                 owner=None,
                 reference_value=None,
                 default_allocation=defaultGatingAllocation,
                 size=None,
                 index=None,
                 assign=None,
                 function=Linear(),
                 modulation:tc.optional(str)=None,
                 modulates=None,
                 params=None,
                 name=None,
                 prefs:is_pref_set=None,
                 **kwargs):

        if index is None and owner is not None:
            allocation = owner.gating_allocation
            if len(allocation)==1:
                index = PRIMARY
            else:
                index = SEQUENTIAL

        # Assign args to params and functionParams dicts
        params = self._assign_args_to_param_dicts(function=function,
                                                  params=params)

        # FIX: 5/26/16
        # IMPLEMENTATION NOTE:
        # Consider adding self to owner.output_states here (and removing from GatingProjection._instantiate_sender)
        #  (test for it, and create if necessary, as per OutputStates in GatingProjection._instantiate_sender),

        # Validate sender (as variable) and params, and assign to variable and paramInstanceDefaults
        super().__init__(owner=owner,
                         reference_value=reference_value,
                         default_allocation=default_allocation,
                         size=size,
                         modulation=modulation,
                         index=index,
                         assign=assign,
                         modulates=modulates,
                         params=params,
                         name=name,
                         prefs=prefs,
                         function=function,
                         **kwargs)

    def _parse_state_specific_specs(self, owner, state_dict, state_specific_spec):
            """Get connections specified in a ParameterState specification tuple

            Tuple specification can be:
                (State name, Mechanism)
            [TBI:] (Mechanism, State name, weight, exponent, projection_specs)

            Returns params dict with CONNECTIONS entries if any of these was specified.

            """
            from psyneulink.core.components.projections.projection import _parse_connection_specs

            params_dict = {}
            state_spec = state_specific_spec

            if isinstance(state_specific_spec, dict):
                return None, state_specific_spec

            elif isinstance(state_specific_spec, tuple):
                state_spec = None
                params_dict[PROJECTIONS] = _parse_connection_specs(connectee_state_type=self,
                                                                   owner=owner,
                                                                   connections=state_specific_spec)
            elif state_specific_spec is not None:
                raise GatingSignalError("PROGRAM ERROR: Expected tuple or dict for {}-specific params but, got: {}".
                                      format(self.__class__.__name__, state_specific_spec))

            if params_dict[PROJECTIONS] is None:
                raise GatingSignalError("PROGRAM ERROR: No entry found in {} params dict for {} "
                                         "with specification of {}, {} or GatingProjection(s) to it".
                                            format(GATING_SIGNAL, INPUT_STATE, OUTPUT_STATE, owner.name))
            return state_spec, params_dict

    def _get_primary_state(self, mechanism):
        return mechanism.input_state

    @property
    def gating_signal(self):
        return self.value



# Princeton University licenses this file to You under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may obtain a copy of the License at:
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.


# ********************************************* LLVM bindings **************************************************************

import ast
import atexit
import ctypes
import functools
import inspect
from llvmlite import ir
import numpy as np
import os
import re
import weakref
try:
    import torch
    torch_available = True
except ImportError:
    torch_available = False

from psyneulink.core.scheduling.time import TimeScale
from psyneulink.core.globals.keywords import AFTER, BEFORE

from psyneulink.core import llvm as pnlvm
from .debug import debug_env
from .helpers import ConditionGenerator

__all__ = ['LLVMBuilderContext', '_modules', '_find_llvm_function']


_modules = set()
_all_modules = set()
_struct_count = 0


@atexit.register
def module_count():
    if "stat" in debug_env:
        print("Total LLVM modules: ", len(_all_modules))
        print("Total structures generated: ", _struct_count)


# TODO: Should this be selectable?
_int32_ty = ir.IntType(32)
_float_ty = ir.DoubleType()
_global_context = None

_BUILTIN_PREFIX = "__pnl_builtin_"
_builtin_intrinsics = frozenset(('pow', 'log', 'exp', 'printf'))

class LLVMBuilderContext:
    uniq_counter = 0
    _llvm_generation = 0

    def __init__(self):
        self.int32_ty = _int32_ty
        self.float_ty = _float_ty
        self._modules = []
        self._cache = weakref.WeakKeyDictionary()
        self._learningcache = weakref.WeakKeyDictionary()
    def __enter__(self):
        module = ir.Module(name="PsyNeuLinkModule-" + str(LLVMBuilderContext._llvm_generation))
        self._modules.append(module)
        LLVMBuilderContext._llvm_generation += 1
        return self

    def __exit__(self, e_type, e_value, e_traceback):
        assert len(self._modules) > 0
        module = self._modules.pop()
        _modules.add(module)
        _all_modules.add(module)

    @property
    def module(self):
        assert len(self._modules) > 0
        return self._modules[-1]

    @staticmethod
    def get_global():
        global _global_context
        if _global_context is None:
            _global_context = LLVMBuilderContext()
        return _global_context

    def get_unique_name(self, name: str):
        LLVMBuilderContext.uniq_counter += 1
        name = re.sub(r"[- ()\[\]]", "_", name)
        return name + '_' + str(LLVMBuilderContext.uniq_counter)

    def get_builtin(self, name: str, args=[], function_type=None):
        if name in _builtin_intrinsics:
            return self.get_llvm_function(_BUILTIN_PREFIX + name)
        if name in ('maxnum'):
            function_type = pnlvm.ir.FunctionType(args[0], [args[0], args[0]])
        return self.module.declare_intrinsic("llvm." + name, args, function_type)

    def create_llvm_function(self, args, component, name=None, return_type=ir.VoidType()):
        name = str(component) if name is None else name

        # Builtins are already unique and need to keep their special name
        func_name = name if name.startswith(_BUILTIN_PREFIX) else self.get_unique_name(name)
        func_ty = pnlvm.ir.FunctionType(return_type, args)
        llvm_func = pnlvm.ir.Function(self.module, func_ty, name=func_name)
        llvm_func.attributes.add('argmemonly')
        for a in llvm_func.args:
            if isinstance(a.type, ir.PointerType):
                a.attributes.add('nonnull')

        metadata = self.get_debug_location(llvm_func, component)
        if metadata is not None:
            scope = dict(metadata.operands)["scope"]
            llvm_func.set_metadata("dbg", scope)

        # Create entry block
        block = llvm_func.append_basic_block(name="entry")
        builder = pnlvm.ir.IRBuilder(block)
        builder.debug_metadata = metadata

        return builder

    def gen_llvm_function(self, obj):
        cache = self._cache
        try:
            # HACK: allows for learning bin func and non-learning to differ
            if obj.learning_enabled is True:
                cache = self._learningcache
        except AttributeError as e:
            pass
            
        if obj not in cache:
            cache[obj] = obj._gen_llvm_function()
        return cache[obj]

    def get_llvm_function(self, name):
        try:
            f = self.gen_llvm_function(name)
        except AttributeError:
            f = _find_llvm_function(name, _all_modules | {self.module})
        # Add declaration to the current module
        if f.name not in self.module.globals:
            decl_f = ir.Function(self.module, f.type.pointee, f.name)
            assert decl_f.is_declaration
            return decl_f
        return f

    @staticmethod
    def get_debug_location(func: ir.Function, component):
        if "debug_info" not in debug_env:
            return

        mod = func.module
        path = inspect.getfile(component.__class__) if component is not None else "<pnl_builtin>"
        d_version = mod.add_metadata([ir.IntType(32)(2), "Dwarf Version", ir.IntType(32)(4)])
        di_version = mod.add_metadata([ir.IntType(32)(2), "Debug Info Version", ir.IntType(32)(3)])
        flags = mod.add_named_metadata("llvm.module.flags")
        if len(flags.operands) == 0:
            flags.add(d_version)
            flags.add(di_version)
        cu = mod.add_named_metadata("llvm.dbg.cu")
        di_file = mod.add_debug_info("DIFile", {
            "filename": os.path.basename(path),
            "directory": os.path.dirname(path),
        })
        di_func_type = mod.add_debug_info("DISubroutineType", {
            # None as `null`
            "types": mod.add_metadata([None]),
        })
        di_compileunit = mod.add_debug_info("DICompileUnit", {
            "language": ir.DIToken("DW_LANG_Python"),
            "file": di_file,
            "producer": "PsyNeuLink",
            "runtimeVersion": 0,
            "isOptimized": False,
        }, is_distinct=True)
        cu.add(di_compileunit)
        di_func = mod.add_debug_info("DISubprogram", {
            "name": func.name,
            "file": di_file,
            "line": 0,
            "type": di_func_type,
            "isLocal": False,
            "unit": di_compileunit,
        }, is_distinct=True)
        di_loc = mod.add_debug_info("DILocation", {
            "line": 0, "column": 0, "scope": di_func,
        })
        return di_loc

    def get_input_struct_type(self, component):
        if hasattr(component, '_get_input_struct_type'):
            return component._get_input_struct_type(self)

        default_var = component.defaults.variable
        return self.convert_python_struct_to_llvm_ir(default_var)

    def get_output_struct_type(self, component):
        if hasattr(component, '_get_output_struct_type'):
            return component._get_output_struct_type(self)

        default_val = component.defaults.value
        return self.convert_python_struct_to_llvm_ir(default_val)

    def get_param_struct_type(self, component):
        if hasattr(component, '_get_param_struct_type'):
            return component._get_param_struct_type(self)

        params = component._get_param_values()
        return self.convert_python_struct_to_llvm_ir(params)

    def get_state_struct_type(self, component):
        if hasattr(component, '_get_state_struct_type'):
            return component._get_state_struct_type(self)

        stateful = component._get_state_values()
        return self.convert_python_struct_to_llvm_ir(stateful)

    def get_data_struct_type(self, component):
        if hasattr(component, '_get_data_struct_type'):
            return component._get_data_struct_type(self)

        return ir.LiteralStructType([])

    def get_param_ptr(self, component, builder, params_ptr, param_name):
        idx = self.int32_ty(component._get_param_ids().index(param_name))
        return builder.gep(params_ptr, [self.int32_ty(0), idx],
                           name="ptr_param_{}_{}".format(param_name, component.name))

    def get_state_ptr(self, component, builder, state_ptr, state_name):
        idx = self.int32_ty(component._get_state_ids().index(state_name))
        return builder.gep(state_ptr, [self.int32_ty(0), idx],
                           name="ptr_state_{}_{}".format(state_name, component.name))

    def unwrap_2d_array(self, builder, element):
        if isinstance(element.type.pointee, ir.ArrayType) and isinstance(element.type.pointee.element, ir.ArrayType):
            assert element.type.pointee.count == 1
            return builder.gep(element, [self.int32_ty(0), self.int32_ty(0)])
        return element

    def inject_printf(self, builder, fmt, *args, override_debug=False):
        if "print_values" not in debug_env and not override_debug:
            return
        fmt += "\0"

        int8 = ir.IntType(8)
        stack_save = self.get_builtin("stacksave", [],
                                      ir.FunctionType(int8.as_pointer(), []))
        stack_restore = self.get_builtin("stackrestore", [],
                                         ir.FunctionType(ir.VoidType(), [int8.as_pointer()]))

        old_stack = builder.call(stack_save, [])
        fmt_data = bytearray(fmt.encode("utf8"))

        # Allocate array to ease initialization
        fmt = builder.alloca(ir.ArrayType(int8, len(fmt_data)))
        builder.store(fmt.type.pointee(fmt_data), fmt)
        fmt_ptr = builder.gep(fmt, [self.int32_ty(0), self.int32_ty(0)])

        printf = self.get_builtin("printf")
        builder.call(printf, [fmt_ptr] + list(args))

        builder.call(stack_restore, [old_stack])


    def inject_printf_float_array(self, builder, array, prefix="", suffix="\n", override_debug=False):
        self.inject_printf(builder,prefix,override_debug=override_debug)

        with pnlvm.helpers.array_ptr_loop(builder, array, "print_array_loop") as (b1, i):
            self.inject_printf(b1, "%f ", b1.load(b1.gep(array, [self.int32_ty(0), i])), override_debug=override_debug)

        self.inject_printf(builder,suffix,override_debug=override_debug)

    def gen_autodiffcomp_learning_exec(self,composition,simulation=False):
        composition._build_pytorch_representation(composition.default_execution_id)
        pytorch_model = composition.parameters.pytorch_representation.get(composition.default_execution_id)
        cond_gen = ConditionGenerator(self, composition)
        
        name = 'exec_learning_sim_wrap_' if simulation else 'exec_learning_wrap_'
        name += composition.name
        args = [self.get_state_struct_type(composition).as_pointer(),
                self.get_param_struct_type(composition).as_pointer(),
                self.get_input_struct_type(composition).as_pointer(),
                self.get_data_struct_type(composition).as_pointer(),
                cond_gen.get_condition_struct_type().as_pointer()]
        builder = self.create_llvm_function(args, composition, name)
        llvm_func = builder.function
        
        for a in llvm_func.args:
            a.attributes.add('noalias')
        
        context, params, comp_in, data_arg, cond = llvm_func.args
        pytorch_model._gen_llvm_training_function_body(self, builder, context, params, comp_in, data_arg, cond)
        # Call output CIM

        if "const_params" in debug_env:
            const_params = params.type.pointee(composition._get_param_initializer(None))
            params = builder.alloca(const_params.type, name="const_params_loc")
            builder.store(const_params, params)

        if "alloca_data" in debug_env:
            data = builder.alloca(data_arg.type.pointee)
            data_vals = builder.load(data_arg)
            builder.store(data_vals, data)
        else:
            data = data_arg

        output_cim_w = composition._get_node_wrapper(composition.output_CIM, simulation)
        output_cim_f = self.get_llvm_function(output_cim_w)
        builder.block.name = "invoke_" + output_cim_f.name
        builder.call(output_cim_f, [context, params, comp_in, data, data])

        if "alloca_data" in debug_env:
            data_vals = builder.load(data)
            builder.store(data_vals, data_arg)

        # Bump run counter
        cond_gen.bump_ts(builder, cond, (1, 0, 0))

        builder.ret_void()
        
        return llvm_func

    def gen_autodiffcomp_exec(self,composition,simulation=False):
        """Creates llvm bin execute for autodiffcomp"""
        assert composition.controller is None
       
        composition._build_pytorch_representation(composition.default_execution_id)
        pytorch_model = composition.parameters.pytorch_representation.get(composition.default_execution_id)
        cond_gen = ConditionGenerator(self, composition)
        
        name = 'exec_sim_wrap_' if simulation else 'exec_wrap_'
        name += composition.name
        args = [self.get_state_struct_type(composition).as_pointer(),
                self.get_param_struct_type(composition).as_pointer(),
                self.get_input_struct_type(composition).as_pointer(),
                self.get_data_struct_type(composition).as_pointer(),
                cond_gen.get_condition_struct_type().as_pointer()]
        builder = self.create_llvm_function(args, composition, name)
        llvm_func = builder.function
        
        for a in llvm_func.args:
            a.attributes.add('noalias')
        
        context, params, comp_in, data_arg, cond = llvm_func.args

        if "const_params" in debug_env:
            const_params = params.type.pointee(composition._get_param_initializer(None))
            params = builder.alloca(const_params.type, name="const_params_loc")
            builder.store(const_params, params)

        if "alloca_data" in debug_env:
            data = builder.alloca(data_arg.type.pointee)
            data_vals = builder.load(data_arg)
            builder.store(data_vals, data)
        else:
            data = data_arg
        
        # Call input CIM
        input_cim_w = composition._get_node_wrapper(composition.input_CIM, simulation)
        input_cim_f = self.get_llvm_function(input_cim_w)

        builder.call(input_cim_f, [context, params, comp_in, data, data])

        # Call pytorch internal compiled llvm func
        pytorch_forward_func = self.get_llvm_function(self.gen_llvm_function(pytorch_model).name)
        input_cim_idx = composition._get_node_index(composition.input_CIM)

        model_context = context
        model_params = builder.gep(params,[self.int32_ty(0),
                                        self.int32_ty(2)])

        # Extract the input that should be inserted into the model
        model_input = builder.gep(data,[self.int32_ty(0),
                                        self.int32_ty(0),
                                        self.int32_ty(input_cim_idx)])
        model_output = builder.gep(data,[self.int32_ty(0),
                                         ])
        
        builder.call(pytorch_forward_func,[model_context,model_params,model_input,model_output])
        
        # Call output CIM
        output_cim_w = composition._get_node_wrapper(composition.output_CIM, simulation)
        output_cim_f = self.get_llvm_function(output_cim_w)
        builder.block.name = "invoke_" + output_cim_f.name
        builder.call(output_cim_f, [context, params, comp_in, data, data])

        if "alloca_data" in debug_env:
            data_vals = builder.load(data)
            builder.store(data_vals, data_arg)

        # Bump run counter
        cond_gen.bump_ts(builder, cond, (1, 0, 0))

        builder.ret_void()
        
        return llvm_func

    def gen_composition_exec(self, composition, simulation=False):
        # Create condition generator
        cond_gen = ConditionGenerator(self, composition)

        name = 'exec_sim_wrap_' if simulation else 'exec_wrap_'
        name += composition.name
        args = [self.get_state_struct_type(composition).as_pointer(),
                self.get_param_struct_type(composition).as_pointer(),
                self.get_input_struct_type(composition).as_pointer(),
                self.get_data_struct_type(composition).as_pointer(),
                cond_gen.get_condition_struct_type().as_pointer()]
                
        builder = self.create_llvm_function(args, composition, name)
        llvm_func = builder.function
        for a in llvm_func.args:
            a.attributes.add('noalias')

        state, params, comp_in, data_arg, cond = llvm_func.args

        if "const_params" in debug_env:
            const_params = params.type.pointee(composition._get_param_initializer(None))
            params = builder.alloca(const_params.type, name="const_params_loc")
            builder.store(const_params, params)

        if "alloca_data" in debug_env:
            data = builder.alloca(data_arg.type.pointee)
            data_vals = builder.load(data_arg)
            builder.store(data_vals, data)
        else:
            data = data_arg

        # Call input CIM
        input_cim_w = composition._get_node_wrapper(composition.input_CIM, simulation)
        input_cim_f = self.get_llvm_function(input_cim_w)
        builder.call(input_cim_f, [state, params, comp_in, data, data])

        if simulation is False and composition.enable_controller and \
           composition.controller_mode == BEFORE:
            assert composition.controller is not None
            controller = composition._get_node_wrapper(composition.controller, simulation)
            controller_f = self.get_llvm_function(controller)
            builder.call(controller_f, [state, params, comp_in, data, data])

        # Allocate run set structure
        run_set_type = ir.ArrayType(ir.IntType(1), len(composition.nodes))
        run_set_ptr = builder.alloca(run_set_type, name="run_set")
        builder.store(run_set_type(None), run_set_ptr)

        # Allocate temporary output storage
        output_storage = builder.alloca(data.type.pointee, name="output_storage")

        iter_ptr = builder.alloca(self.int32_ty, name="iter_counter")
        builder.store(self.int32_ty(0), iter_ptr)

        loop_condition = builder.append_basic_block(name="scheduling_loop_condition")
        builder.branch(loop_condition)

        # Generate a while not 'end condition' loop
        builder.position_at_end(loop_condition)
        run_cond = cond_gen.generate_sched_condition(builder,
                        composition.termination_processing[TimeScale.TRIAL],
                        cond, None)
        run_cond = builder.not_(run_cond, name="not_run_cond")

        loop_body = builder.append_basic_block(name="scheduling_loop_body")
        exit_block = builder.append_basic_block(name="exit")
        builder.cbranch(run_cond, loop_body, exit_block)

        # Generate loop body
        builder.position_at_end(loop_body)

        zero = self.int32_ty(0)
        any_cond = ir.IntType(1)(0)

        # Calculate execution set before running the mechanisms
        for idx, mech in enumerate(composition.nodes):
            run_set_mech_ptr = builder.gep(run_set_ptr,
                                           [zero, self.int32_ty(idx)],
                                           name="run_cond_ptr_" + mech.name)
            mech_cond = cond_gen.generate_sched_condition(builder,
                            composition._get_processing_condition_set(mech),
                            cond, mech)
            ran = cond_gen.generate_ran_this_pass(builder, cond, mech)
            mech_cond = builder.and_(mech_cond, builder.not_(ran),
                                     name="run_cond_" + mech.name)
            any_cond = builder.or_(any_cond, mech_cond, name="any_ran_cond")
            builder.store(mech_cond, run_set_mech_ptr)

        for idx, mech in enumerate(composition.nodes):
            run_set_mech_ptr = builder.gep(run_set_ptr, [zero, self.int32_ty(idx)])
            mech_cond = builder.load(run_set_mech_ptr, name="mech_" + mech.name + "_should_run")
            with builder.if_then(mech_cond):
                mech_w = composition._get_node_wrapper(mech, simulation)
                mech_f = self.get_llvm_function(mech_w)
                builder.block.name = "invoke_" + mech_f.name
                # Wrappers do proper indexing of all structures
                if len(mech_f.args) == 5:  # Mechanism wrappers have 5 inputs
                    builder.call(mech_f, [state, params, comp_in, data, output_storage])
                else:
                    builder.call(mech_f, [state, params, comp_in, data, output_storage, cond])

                cond_gen.generate_update_after_run(builder, cond, mech)
            builder.block.name = "post_invoke_" + mech_f.name

        # Writeback results
        for idx, mech in enumerate(composition.nodes):
            run_set_mech_ptr = builder.gep(run_set_ptr, [zero, self.int32_ty(idx)])
            mech_cond = builder.load(run_set_mech_ptr, name="mech_" + mech.name + "_ran")
            with builder.if_then(mech_cond):
                out_ptr = builder.gep(output_storage, [zero, zero, self.int32_ty(idx)], name="result_ptr_" + mech.name)
                data_ptr = builder.gep(data, [zero, zero, self.int32_ty(idx)],
                                       name="data_result_" + mech.name)
                builder.store(builder.load(out_ptr), data_ptr)

        # Update step counter
        with builder.if_then(any_cond):
            builder.block.name = "inc_step"
            cond_gen.bump_ts(builder, cond)

        builder.block.name = "update_iter_count"
        # Increment number of iterations
        iters = builder.load(iter_ptr, name="iterw")
        iters = builder.add(iters, self.int32_ty(1), name="iterw_inc")
        builder.store(iters, iter_ptr)

        max_iters = len(composition.scheduler_processing.consideration_queue)
        completed_pass = builder.icmp_unsigned("==", iters,
                                               self.int32_ty(max_iters),
                                               name="completed_pass")
        # Increment pass and reset time step
        with builder.if_then(completed_pass):
            builder.block.name = "inc_pass"
            builder.store(zero, iter_ptr)
            # Bumping automatically zeros lower elements
            cond_gen.bump_ts(builder, cond, (0, 1, 0))

        builder.branch(loop_condition)

        builder.position_at_end(exit_block)

        if simulation is False and composition.enable_controller and \
           composition.controller_mode == AFTER:
            assert composition.controller is not None
            controller = composition._get_node_wrapper(composition.controller, simulation)
            controller_f = self.get_llvm_function(controller)
            builder.call(controller_f, [state, params, comp_in, data, data])

        # Call output CIM
        output_cim_w = composition._get_node_wrapper(composition.output_CIM, simulation)
        output_cim_f = self.get_llvm_function(output_cim_w)
        builder.block.name = "invoke_" + output_cim_f.name
        builder.call(output_cim_f, [state, params, comp_in, data, data])

        if "alloca_data" in debug_env:
            data_vals = builder.load(data)
            builder.store(data_vals, data_arg)

        # Bump run counter
        cond_gen.bump_ts(builder, cond, (1, 0, 0))

        builder.ret_void()

        return llvm_func

    def gen_composition_run(self, composition, simulation=False):
        name = 'run_sim_wrap_' if simulation else 'run_wrap_'
        name += composition.name
        args = [self.get_state_struct_type(composition).as_pointer(),
                self.get_param_struct_type(composition).as_pointer(),
                self.get_data_struct_type(composition).as_pointer(),
                self.get_input_struct_type(composition).as_pointer(),
                self.get_output_struct_type(composition).as_pointer(),
                self.int32_ty.as_pointer(),
                self.int32_ty.as_pointer()]
        builder = self.create_llvm_function(args, composition, name)
        llvm_func = builder.function
        for a in llvm_func.args:
            a.attributes.add('noalias')

        state, params, data, data_in, data_out, runs_ptr, inputs_ptr = llvm_func.args
        # simulation does not care about the output
        # it extracts results of the controller objective mechanism
        if simulation:
            data_out.attributes.remove('nonnull')

        if not simulation and "const_data" in debug_env:
            const_data = data.type.pointee(composition._get_data_initializer(None))
            data = builder.alloca(data.type.pointee)
            builder.store(const_data, data)

        # Hardcode stateful parameters if set in the environment
        if not simulation and "const_state" in debug_env:
            const_state = state.type.pointee(composition._get_state_initializer(None))
            state = builder.alloca(const_state.type, name="const_state_loc")
            builder.store(const_state, state)

        if not simulation and "const_input" in debug_env:
            if not debug_env["const_input"]:
                input_init = pnlvm._tupleize([[os.defaults.variable] for os in composition.input_CIM.input_states])
                print("Setting default input: ", input_init)
            else:
                input_init = ast.literal_eval(debug_env["const_input"])
                print("Setting user input: ", input_init)

            builder.store(data_in.type.pointee(input_init), data_in)
            builder.store(inputs_ptr.type.pointee(1), inputs_ptr)

        if "force_runs" in debug_env:
            num = int(debug_env["force_runs"]) if debug_env["force_runs"] else 1
            print("Forcing number of runs to: ", num)
            runs_ptr = builder.alloca(runs_ptr.type.pointee)
            builder.store(runs_ptr.type.pointee(num), runs_ptr)

        # Allocate and initialize condition structure
        cond_gen = ConditionGenerator(self, composition)
        cond_type = cond_gen.get_condition_struct_type()
        cond = builder.alloca(cond_type)
        cond_init = cond_type(cond_gen.get_condition_initializer())
        builder.store(cond_init, cond)

        iter_ptr = builder.alloca(self.int32_ty, name="iter_counter")
        builder.store(self.int32_ty(0), iter_ptr)

        loop_condition = builder.append_basic_block(name="run_loop_condition")
        builder.branch(loop_condition)

        # Generate a "while < count" loop
        builder.position_at_end(loop_condition)
        count = builder.load(iter_ptr)
        runs = builder.load(runs_ptr)
        run_cond = builder.icmp_unsigned('<', count, runs)

        loop_body = builder.append_basic_block(name="run_loop_body")
        exit_block = builder.append_basic_block(name="exit")
        builder.cbranch(run_cond, loop_body, exit_block)

        # Generate loop body
        builder.position_at_end(loop_body)

        # Current iteration
        iters = builder.load(iter_ptr)

        # Get the right input stimulus
        input_idx = builder.urem(iters, builder.load(inputs_ptr))
        data_in_ptr = builder.gep(data_in, [input_idx])
        
        # Call execution
        if simulation:
            exec_f = self.get_llvm_function(composition._llvm_simulation.name)
        else:
            exec_f = self.get_llvm_function(composition)
        builder.call(exec_f, [state, params, data_in_ptr, data, cond])

        if not simulation:
            # Extract output_CIM result
            idx = composition._get_node_index(composition.output_CIM)
            result_ptr = builder.gep(data, [self.int32_ty(0), self.int32_ty(0), self.int32_ty(idx)])
            output_ptr = builder.gep(data_out, [iters])
            result = builder.load(result_ptr)
            builder.store(result, output_ptr)

        # Increment counter
        iters = builder.add(iters, self.int32_ty(1))
        builder.store(iters, iter_ptr)
        builder.branch(loop_condition)

        builder.position_at_end(exit_block)

        # Store the number of executed iterations
        builder.store(builder.load(iter_ptr), runs_ptr)

        builder.ret_void()

        return llvm_func

    def gen_multirun_wrapper(self, function: ir.Function) -> ir.Function:
        if function.module is not self.module:
            function = ir.Function(self.module, function.type.pointee, function.name)
            assert function.is_declaration

        args = [a.type for a in function.args]
        args.append(self.int32_ty.as_pointer())
        multirun_ty = ir.FunctionType(function.type.pointee.return_type, args)
        multirun_f = ir.Function(self.module, multirun_ty, function.name + "_multirun")
        block = multirun_f.append_basic_block(name="entry")
        cond_block = multirun_f.append_basic_block(name="loop_cond")
        body_block = multirun_f.append_basic_block(name="loop_body")
        exit_block = multirun_f.append_basic_block(name="exit_loop")

        builder = ir.IRBuilder(block)

        limit_ptr = multirun_f.args[-1]
        index_ptr = builder.alloca(self.int32_ty)
        builder.store(index_ptr.type.pointee(0), index_ptr)
        builder.branch(cond_block)

        with builder.goto_block(cond_block):
            index = builder.load(index_ptr)
            limit = builder.load(limit_ptr)
            cond = builder.icmp_unsigned("<", index, limit)
            builder.cbranch(cond, body_block, exit_block)

        with builder.goto_block(body_block):
            # Runs need special handling. data_in and data_out are one dimensional,
            # but hold entries for all parallel invocations.
            is_comp_run = len(function.args) == 7
            if is_comp_run:
                runs_count = multirun_f.args[5]
                input_count = multirun_f.args[6]

            # Index all pointer arguments
            index = builder.load(index_ptr)
            indexed_args = []
            for i, arg in enumerate(multirun_f.args[:-1]):
                # Don't adjust #inputs and #trials
                if isinstance(arg.type, ir.PointerType):
                    offset = index
                    # #runs and #trials needs to be the same
                    if is_comp_run and i >= 5:
                        offset = self.int32_ty(0)
                    # data arrays need special handling
                    elif is_comp_run and i == 4:  # data_out
                        offset = builder.mul(index, builder.load(runs_count))
                    elif is_comp_run and i == 3:  # data_in
                        offset = builder.mul(index, builder.load(input_count))

                    arg = builder.gep(arg, [offset])

                indexed_args.append(arg)

            builder.call(function, indexed_args)
            new_idx = builder.add(index, index.type(1))
            builder.store(new_idx, index_ptr)
            builder.branch(cond_block)

        with builder.goto_block(exit_block):
            builder.ret_void()

        return multirun_f

    def convert_python_struct_to_llvm_ir(self, t):
        if type(t) is list:
            assert all(type(x) is type(t[0]) for x in t)
            elem_t = self.convert_python_struct_to_llvm_ir(t[0])
            return ir.ArrayType(elem_t, len(t))
        elif type(t) is tuple:
            elems_t = (self.convert_python_struct_to_llvm_ir(x) for x in t)
            return ir.LiteralStructType(elems_t)
        elif isinstance(t, (int, float)):
            return self.float_ty
        elif isinstance(t, np.ndarray):
            return self.convert_python_struct_to_llvm_ir(t.tolist())
        elif t is None:
            return ir.LiteralStructType([])
        elif isinstance(t, np.random.RandomState):
            return pnlvm.builtins.get_mersenne_twister_state_struct(self)
        elif torch_available and isinstance(t,torch.Tensor):
            return self.convert_python_struct_to_llvm_ir(t.numpy())
        assert False, "Don't know how to convert {}".format(type(t))


def _find_llvm_function(name: str, mods=_all_modules) -> ir.Function:
    f = None
    for m in mods:
        if name in m.globals:
            f = m.get_global(name)

    if not isinstance(f, ir.Function):
        raise ValueError("No such function: {}".format(name))
    return f


def _gen_cuda_kernel_wrapper_module(function):
    module = ir.Module(name="wrapper_" + function.name)

    decl_f = ir.Function(module, function.type.pointee, function.name)
    assert decl_f.is_declaration
    kernel_func = ir.Function(module, function.type.pointee, function.name + "_cuda_kernel")
    block = kernel_func.append_basic_block(name="entry")
    builder = ir.IRBuilder(block)

    # Calculate global id of a thread in x dimension
    intrin_ty = ir.FunctionType(ir.IntType(32), [])
    tid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.tid.x")
    ntid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.ntid.x")
    ctaid_x_f = ir.Function(module, intrin_ty, "llvm.nvvm.read.ptx.sreg.ctaid.x")
    global_id = builder.mul(builder.call(ctaid_x_f, []), builder.call(ntid_x_f, []))
    global_id = builder.add(global_id, builder.call(tid_x_f, []))

    # Runs need special handling. data_in and data_out are one dimensional,
    # but hold entries for all parallel invocations.
    is_comp_run = len(kernel_func.args) == 7
    if is_comp_run:
        runs_count = kernel_func.args[5]
        input_count = kernel_func.args[6]

    # Index all pointer arguments
    indexed_args = []
    for i, arg in enumerate(kernel_func.args):
        # Don't adjust #inputs and #trials
        if isinstance(arg.type, ir.PointerType):
            offset = global_id
            # #runs and #trials needs to be the same
            if is_comp_run and i >= 5:
                offset = ir.IntType(32)(0)
            # data arrays need special handling
            elif is_comp_run and i == 4:  # data_out
                offset = builder.mul(global_id, builder.load(runs_count))
            elif is_comp_run and i == 3:  # data_in
                offset = builder.mul(global_id, builder.load(input_count))

            arg = builder.gep(arg, [offset])

        indexed_args.append(arg)
    builder.call(decl_f, indexed_args)
    builder.ret_void()

    # Add kernel mark metadata
    module.add_named_metadata("nvvm.annotations", [kernel_func, "kernel", ir.IntType(32)(1)])

    return module


@functools.lru_cache(maxsize=128)
def _convert_llvm_ir_to_ctype(t):
    type_t = type(t)

    if type_t is ir.VoidType:
        return None
    elif type_t is ir.IntType:
        if t.width == 32:
            return ctypes.c_int
        elif t.width == 64:
            return ctypes.c_longlong
        else:
            assert False, "Integer type too big!"
    elif type_t is ir.DoubleType:
        return ctypes.c_double
    elif type_t is ir.FloatType:
        return ctypes.c_float
    elif type_t is ir.PointerType:
        # FIXME: Can this handle void*? Do we care?
        pointee = _convert_llvm_ir_to_ctype(t.pointee)
        ret_t = ctypes.POINTER(pointee)
    elif type_t is ir.ArrayType:
        element_type = _convert_llvm_ir_to_ctype(t.element)
        ret_t = element_type * len(t)
    elif type_t is ir.LiteralStructType:
        global _struct_count
        uniq_name = "struct_" + str(_struct_count)
        _struct_count += 1

        field_list = []
        for i, e in enumerate(t.elements):
            # llvmlite modules get _unique string only works for symbol names
            field_uniq_name = uniq_name + "field_" + str(i)
            field_list.append((field_uniq_name, _convert_llvm_ir_to_ctype(e)))

        ret_t = type(uniq_name, (ctypes.Structure,), {"__init__": ctypes.Structure.__init__})
        ret_t._fields_ = field_list
        assert len(ret_t._fields_) == len(t.elements)
    else:
        assert False, "Don't know how to convert LLVM type: {}".format(t)

    return ret_t



import numpy as np
import itertools


def generate_inputs(num_dimensions, num_features):
    base = np.zeros((1, num_features))
    base[0, 0] = 1.0

    base_perm = base.copy()  # make a copy as to not modify the base vector

    # roll the vector to get all possible nodes active in a single input dimension
    for i in range(1, num_features):
        base_perm = np.vstack((base_perm, np.roll(base, i, axis=1)))

    # get all possible combinations of input dimensions
    return np.array([np.array(p) for p in itertools.product(base_perm, repeat=num_dimensions)])


def generate_tasks(num_dimensions, out_dimensions=None):
    if out_dimensions is None:
        out_dimensions = num_dimensions

    units = num_dimensions * out_dimensions

    # generate base vector
    base = np.zeros((1, units))
    base[0, 0] = 1.0

    tasks = base.copy()

    # roll vector to get all possible nodes active
    for i in range(1, units):
        tasks = np.vstack((tasks, np.roll(base, i, axis=1)))

    # Generate a map of dimensions mapped by each task
    task_numbers = np.arange(units)
    task_map = np.reshape(task_numbers, (num_dimensions, out_dimensions))

    return tasks, task_map


def generate_training_patterns(num_dimensions, num_features, inputs=None, tasks=None, task_map=None):
    if inputs is None:
        inputs = generate_inputs(num_dimensions, num_features)

    if tasks is None or task_map is None:
        tasks, task_map = generate_tasks(num_dimensions)

    num_cases = len(inputs) * len(tasks)

    # changed to a tensor to support the sort of indexing I need
    target_patterns = np.zeros((num_cases, num_dimensions, num_features))
    input_patterns = np.zeros((num_cases, num_dimensions, num_features))
    task_patterns = np.zeros((num_cases, np.size(tasks, 1)))
    in_out_map = np.zeros((num_cases, task_map.ndim))  # logs dim map per trial

    # generate input-task combination and target pattern per trial
    trial = 0

    for task_p in tasks:  # loops along task patterns
        for input_p in inputs:  # loops along input patterns
            input_patterns[trial] = input_p
            task_patterns[trial] = task_p

            task_index = np.argmax(task_p) # np.where(task_p == np.max(task_p))[0][0]

            mapping = np.array(np.where(task_map == task_index))[:,0]
            in_out_map[trial] = mapping
            in_dim = mapping[0]
            out_dim = mapping[1]

            # Perform the dimension mapping from input to output
            target_patterns[trial, out_dim, :] = input_patterns[trial, in_dim, :]

            trial += 1

    return input_patterns, task_patterns, in_out_map, target_patterns


if __name__ == '__main__':
    # ins = generate_inputs(3, 4)
    # print(ins[0])
    # print(ins[0][0])
    # print(ins[0].shape)

    # input_patterns, task_patterns, in_out_map, target_patterns = generate_training_patterns(3, 4)
    # x = 333
    # print(input_patterns[x])
    # print(task_patterns[x])
    # print(in_out_map[x])
    # print(target_patterns[x])
    # print(np.argmax(task_patterns, 1))
    pass






import pytest
import numpy as np
import psyneulink as pnl
import psyneulink.core.components.functions.transferfunctions


def test_control_mechanism_assignment():
    """ControlMechanism assignment/replacement,  monitor_for_control, and control_signal specifications"""

    T1 = pnl.TransferMechanism(size=3, name='T-1')
    T2 = pnl.TransferMechanism(function=psyneulink.core.components.functions.transferfunctions.Logistic, output_states=[{pnl.NAME: 'O-1'}], name='T-2')
    T3 = pnl.TransferMechanism(function=psyneulink.core.components.functions.transferfunctions.Logistic, name='T-3')
    T4 = pnl.TransferMechanism(function=psyneulink.core.components.functions.transferfunctions.Logistic, name='T-4')
    P = pnl.Process(pathway=[T1, T2, T3, T4])
    S = pnl.System(processes=P,
                   # controller=pnl.EVCControlMechanism,
                   controller=pnl.EVCControlMechanism(
                           control_signals=[(pnl.GAIN, T2)],
                           monitor_for_control=T4
                   ),
                   enable_controller=True,
                   # Test for use of 4-item tuple with matrix in monitor_for_control specification
                   monitor_for_control=[(T1, None, None, np.ones((3,1))),
                                        ('O-1', 1, -1)],
                   control_signals=[(pnl.GAIN, T3)]
                   )
    assert len(S.controller.objective_mechanism.monitor)==3
    assert len(S.control_signals)==2

    # Test for avoiding duplicate assignment of monitor and control_signals
    C1 = pnl.EVCControlMechanism(name='C-1',
                                 objective_mechanism = [(T1, None, None, np.ones((3,1)))],
                                 control_signals=[(pnl.GAIN, T3)]
                                 )

    # Test direct assignment
    S.controller = C1
    assert len(C1.monitored_output_states)==2
    assert len(S.control_signals)==3
    assert S.controller.name == 'C-1'


    # Test for adding a monitored_output_state and control_signal
    C2 = pnl.EVCControlMechanism(name='C-2',
                                 objective_mechanism = [T3.output_states[pnl.RESULTS]],
                                 control_signals=[(pnl.GAIN, T4)])
    # Test use of assign_as_controller method
    C2.assign_as_controller(S)
    assert len(C2.monitored_output_states)==3
    assert len(S.control_signals)==4
    assert S.controller.name == 'C-2'


# def test_control_mechanism_assignment_additional():
#     """Tests "free-standing" specifications of monitor_for_control and ControlSignal (i.e., outside of a list)"""
#     T_1 = pnl.TransferMechanism(name='T_1')
#     T_2 = pnl.TransferMechanism(name='T_2')
#     T_3 = pnl.TransferMechanism(name='T_3')
#     S = pnl.sys([T_1, T_2, T_3],
#                 controller=pnl.EVCControlMechanism(
#                         control_signals=(pnl.SLOPE, T_1),
#                         monitor_for_control=[T_1],
#                         objective_mechanism=[T_2]),
#                 monitor_for_control=T_3,
#                 control_signals=(pnl.SLOPE, T_2),
#                 enable_controller=True)
#     assert S.controller.objective_mechanism.input_state[0].path_afferents[0].sender.owner == T_1
#     assert S.controller.objective_mechanism.input_state[1].path_afferents[0].sender.owner == T_2
#     assert S.controller.objective_mechanism.input_state[2].path_afferents[0].sender.owner == T_3
#     assert T_1.parameter_states[pnl.SLOPE].mod_afferents[0].sender.owner == S.controller
#     assert T_2.parameter_states[pnl.SLOPE].mod_afferents[0].sender.owner == S.controller



def test_control_mechanism_assignment_additional():
    """Tests "free-standing" specifications of monitor_for_control and ControlSignal (i.e., outside of a list)"""
    T_1 = pnl.TransferMechanism(name='T_1')
    T_2 = pnl.TransferMechanism(name='T_2')
    S = pnl.sys([T_1,T_2],
                controller=pnl.EVCControlMechanism(control_signals=(pnl.SLOPE, T_1)),
                monitor_for_control=[T_1],
                control_signals=(pnl.SLOPE, T_2),
                enable_controller=True)
    assert S.controller.objective_mechanism.input_state.path_afferents[0].sender.owner == T_1
    assert T_1.parameter_states[pnl.SLOPE].mod_afferents[0].sender.owner == S.controller
    assert T_2.parameter_states[pnl.SLOPE].mod_afferents[0].sender.owner == S.controller

def test_prediction_mechanism_assignment():
    """Tests prediction mechanism assignment and more tests for ObjectiveMechanism and ControlSignal assignments"""

    T1 = pnl.TransferMechanism(name='T1')
    T2 = pnl.TransferMechanism(name='T2')
    T3 = pnl.TransferMechanism(name='T3')

    S = pnl.sys([T1, T2, T3],
                # controller=pnl.EVCControlMechanism(name='EVC',
                controller=pnl.EVCControlMechanism(name='EVC',
                                                   prediction_mechanisms=(pnl.PredictionMechanism,
                                                                          {pnl.FUNCTION: pnl.INPUT_SEQUENCE,
                                                                           pnl.RATE: 1,
                                                                           pnl.WINDOW_SIZE: 3,
                                                                           }),
                                                   monitor_for_control=[T1],
                                                   objective_mechanism=[T2]
                                                   ),
                control_signals=pnl.ControlSignal(allocation_samples=[1, 5, 10],
                                                  modulates=(pnl.SLOPE, T1)),
                monitor_for_control=T3,
                enable_controller=True
                )
    assert len(S.controller.objective_mechanism.input_states)==3

    S.recordSimulationPref = True

    input_dict = {T1:[1,2,3,4]}
    results = S.run(inputs=input_dict)
    assert results == [[[1.]], [[2.]], [[15.]], [[20.]]]
    assert S.simulation_results ==  [[[1.]], [[5.]], [[10.]],
                                    [[1.]], [[2.]], [[5.]], [[10.]], [[10.]], [[20.]],
                                    [[1.]], [[2.]], [[3.]], [[5.]], [[10.]], [[15.]], [[10.]], [[20.]], [[30.]],
                                    [[2.]], [[3.]], [[4.]], [[10.]], [[15.]], [[20.]], [[20.]], [[30.]], [[40.]]]

def test_prediction_mechanism_filter_function():
    """Tests prediction mechanism assignment and more tests for ObjectiveMechanism and ControlSignal assignments"""

    f = lambda x: [x[0]*7]
    T = pnl.TransferMechanism(name='T')

    S = pnl.sys(T,
                controller=pnl.EVCControlMechanism(name='EVC',
                                                   prediction_mechanisms=(pnl.PredictionMechanism,
                                                                          {pnl.FUNCTION: pnl.INPUT_SEQUENCE,
                                                                           pnl.RATE: 1,
                                                                           pnl.WINDOW_SIZE: 3,
                                                                           pnl.FILTER_FUNCTION: f
                                                                           }),
                                                   objective_mechanism=[T]
                                                   ),
                control_signals=pnl.ControlSignal(allocation_samples=[1, 5, 10],
                                                  modulates=(pnl.SLOPE, T)),
                enable_controller=True
                )

    S.recordSimulationPref = True
    input_dict = {T: [1, 2, 3, 4]}
    results = S.run(inputs=input_dict)
    expected_results = [[[1.0]], [[2.0]], [[3.0]], [[4.0]]]
    expected_sim_results = [[[1.]], [[5.]], [[10.]],    # before EVC | [1]
                            [[7.]], [[35.]], [[70.]],   # [1, 2]
                            [[7.]], [[35.]], [[70.]],   # [1, 2, 3]
                            [[14.]], [[70.]], [[140.]]] # [2, 3, 4]

    np.testing.assert_allclose(results, expected_results, atol=1e-08, err_msg='Failed on results')
    np.testing.assert_allclose(S.simulation_results, expected_sim_results, atol=1e-08, err_msg='Failed on results')



from . import adaptivemechanism
from . import modulatorymechanism
from . import control
from . import gating
from . import learning

from .adaptivemechanism import *
from .modulatorymechanism import *
from .control import *
from .gating import *
from .learning import *

__all__ = list(control.__all__)
__all__.extend(gating.__all__)
__all__.extend(learning.__all__)
__all__.extend(adaptivemechanism.__all__)
__all__.extend(modulatorymechanism.__all__)


import pytest

import doctest
import psyneulink as pnl

def test_state_docs():
    # get examples of mechanisms that can be used with GatingSignals/Mechanisms
    pass


def test_parameter_state_docs():
    fail, total = doctest.testmod(pnl.core.components.states.parameterstate, globs={})

    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total),
                    pytrace=False)


def test_output_state_docs():
    fail, total = doctest.testmod(pnl.core.components.states.outputstate)

    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total))


def test_control_signal_docs():
    fail, total = doctest.testmod(pnl.core.components.states.modulatorysignals.controlsignal)

    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total))


def test_gating_signal_docs():
    fail, total = doctest.testmod(pnl.core.components.states.modulatorysignals.gatingsignal)

    if fail > 0:
        pytest.fail("{} out of {} examples failed".format(fail, total))
